"""Module for parsing CSAF v2 advisories"""
import json
from dataclasses import dataclass, InitVar, field
from typing import Any, Iterable

class RemediationParseError(ValueError):
    pass

@dataclass
class Remediation:
    """
    class to handle remediation advice in CSAF data
    """

    csaf_product_id: InitVar[str]
    cpes: InitVar[dict[str, str]]
    purls: InitVar[dict[str, str]]
    product: str = field(init=False)
    product_version: str = field(init=False)
    component: str = field(init=False)
    fixed_version: str = field(init=False)
    purl: str = field(init=False)
    cpe: str = field(init=False)

    def __post_init__(self, csaf_product_id: str, cpes: dict[str, str],
                      purls: dict[str, str]):
        if ":" not in csaf_product_id:
            raise ValueError(
                f"Did not find ':' in product_id: {csaf_product_id}")
        (self.product, self.product_version) = csaf_product_id.split(":",
                                                                     maxsplit=1)

        # NEVRA stands for Name Epoch Version Release and Architecture
        # We split the name from the rest of the 'version' data (EVRA). We store name as component.
        split_component_version = self.product_version.rsplit("-", maxsplit=2)
        if len(split_component_version) < 3:
            raise RemediationParseError(
                f"Could not convert component into NEVRA: {self.product_version}"
            )
        # RHEL Modules have 4 colons in the name part of the NEVRA. If we detect a modular RPM
        # product ID, discard the module part of the name and look for that in the purl dict.
        # Ideally we would keep the module information and use it when scanning a RHEL system,
        # however this is not done today by Clair:  https://github.com/quay/claircore/pull/901/files
        if split_component_version[0].count(":") == 4:
            self.component = split_component_version[0].rsplit(":")[-1]
        else:
            self.component = split_component_version[0]
        self.fixed_version = "-".join(
            (split_component_version[1], split_component_version[2]))

        try:
            nevra = f"{self.component}-{self.fixed_version}"
            self.purl = purls[nevra]
            self.cpe = cpes[self.product]
        except KeyError:
            # pylint: disable=raise-missing-from
            # Raising this as a ValueError instead of as a KeyError allows us to wrap
            # the entire call to init() in try/catch block with a single exception type
            raise ValueError(
                f"Did not find {csaf_product_id} in product branches")

        # There are many pkg:oci/ remediations in Red Hat data. However there are no strict
        # rules enforced on versioning Red Hat containers, therefore we cant compare container
        # versions to each other with 100% accuracy at this time.
        if not self.purl.startswith("pkg:rpm/"):
            raise ValueError(
                "Non RPM remediations are not supported in OSV at this time")


@dataclass
class Vulnerability:
    """
    class to handle vulnerability information
    """

    csaf_vuln: InitVar[dict[str, Any]]
    cpes: InitVar[dict[str, str]]
    purls: InitVar[dict[str, str]]
    cve_id: str = field(init=False)
    cvss_v3_vector: str = field(init=False)
    cvss_v3_base_score: str = field(init=False, default=None)
    references: list[dict[str, str]] = field(init=False)
    remediations: list[Remediation] = field(init=False)

    def __post_init__(self, csaf_vuln: dict[str, Any], cpes: dict[str, str],
                      purls: dict[str, str]):
        self.cve_id = csaf_vuln["cve"]
        if not hasattr(csaf_vuln, "scores"):
            self.cvss_v3_vector = ""
            self.cvss_v3_base_score = ""
        for score in csaf_vuln.get("scores", []):
            if "cvss_v3" in score:
                self.cvss_v3_vector = score["cvss_v3"]["vectorString"]
                self.cvss_v3_base_score = score["cvss_v3"]["baseScore"]
            else:
                self.cvss_v3_base_score = ""
                self.cvss_v3_vector = ""
        self.references = csaf_vuln["references"]
        self.remediations = []
        for product_id in csaf_vuln["product_status"]["fixed"]:
            try:
                self.remediations.append(Remediation(product_id, cpes, purls))
            except RemediationParseError:
                continue
        if not self.remediations:
            raise ValueError(f"Did not find any remediations for {self.cve_id}")


def gen_dict_extract(key, var: Iterable):
    """
    Given a key value and dictionary or list, traverses that dictionary or list returning the value
    of the given key.
    From https://stackoverflow.com/questions/9807634/
        find-all-occurrences-of-a-key-in-nested-dictionaries-and-lists
    """
    if hasattr(var, "items"):
        for k, v in var.items():
            if k == key:
                yield v
            if isinstance(v, dict):
                yield from gen_dict_extract(key, v)
            elif isinstance(v, list):
                for d in v:
                    yield from gen_dict_extract(key, d)


def build_product_maps(
        product_tree_branches: dict) -> tuple[dict[str, str], dict[str, str]]:
    """
    Given a CSAF product tree branch dictionary returns a tuple of CPEs by product ID and PURLs by
    product ID.
    """
    cpe_map = {}
    purl_map = {}
    products = gen_dict_extract("product", product_tree_branches)
    for product in products:
        product_id = product["product_id"]
        if "product_identification_helper" in product:
            helper = product["product_identification_helper"]
            if "cpe" in helper:
                cpe_map[product_id] = helper["cpe"]
            elif "purl" in helper:
                purl_map[product_id] = helper["purl"]
    return cpe_map, purl_map


class CSAF:
    """
    class to handle CSAF data read from a local file path
    """

    def __init__(self, csaf_content: str):
        csaf_data = json.loads(csaf_content)

        if not csaf_data:
            raise ValueError("Unable to load CSAF JSON data.")

        self.doc = csaf_data["document"]

        self.csaf = {
            "type": self.doc["category"],
            "csaf_version": self.doc["csaf_version"]
        }

        # Only support csaf_vex 2.0
        if self.csaf != {"type": "csaf_security_advisory", "csaf_version": "2.0"}:
            raise ValueError(
                f"Can only handle csaf_security_advisory 2.0 documents. Got: {self.csaf}")

        self.cpes, self.purls = build_product_maps(csaf_data["product_tree"])

        self.vulnerabilities = [
            Vulnerability(v, self.cpes, self.purls)
            for v in (csaf_data["vulnerabilities"])
        ]

    @property
    def title(self):
        """
        Document Title
        """
        return self.doc["title"]

    @property
    def references(self):
        """
        Document References
        """
        return self.doc["references"]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Miscellaneous utility functions."""

import os
import google.auth
from google.auth.exceptions import DefaultCredentialsError

_google_cloud_project = None


def get_google_cloud_project() -> str:
  """Determine the current Google Cloud Project.
  
  returns an empty string if project could not be determined.
  """
  global _google_cloud_project
  if _google_cloud_project is not None:
    return _google_cloud_project

  # google.auth.default will also check this env var, but this is cheaper.
  _google_cloud_project = os.getenv('GOOGLE_CLOUD_PROJECT')
  if _google_cloud_project:
    return _google_cloud_project

  try:
    _, _google_cloud_project = google.auth.default()
    if _google_cloud_project:
      return _google_cloud_project
  except DefaultCredentialsError:
    pass

  _google_cloud_project = ''
  return _google_cloud_project



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _ImportFindingType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ImportFindingTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ImportFindingType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    IMPORT_FINDING_TYPE_NONE: _ImportFindingType.ValueType  # 0
    IMPORT_FINDING_TYPE_DELETED: _ImportFindingType.ValueType  # 1
    IMPORT_FINDING_TYPE_INVALID_JSON: _ImportFindingType.ValueType  # 2
    IMPORT_FINDING_TYPE_INVALID_PACKAGE: _ImportFindingType.ValueType  # 3
    IMPORT_FINDING_TYPE_INVALID_PURL: _ImportFindingType.ValueType  # 4
    IMPORT_FINDING_TYPE_INVALID_VERSION: _ImportFindingType.ValueType  # 5
    IMPORT_FINDING_TYPE_INVALID_COMMIT: _ImportFindingType.ValueType  # 6
    IMPORT_FINDING_TYPE_INVALID_RANGE: _ImportFindingType.ValueType  # 7
    IMPORT_FINDING_TYPE_INVALID_RECORD: _ImportFindingType.ValueType  # 8
    IMPORT_FINDING_TYPE_INVALID_ALIASES: _ImportFindingType.ValueType  # 9
    IMPORT_FINDING_TYPE_INVALID_UPSTREAM: _ImportFindingType.ValueType  # 10
    IMPORT_FINDING_TYPE_INVALID_RELATED: _ImportFindingType.ValueType  # 11
    IMPORT_FINDING_TYPE_BAD_ALIASED_CVE: _ImportFindingType.ValueType  # 12

class ImportFindingType(_ImportFindingType, metaclass=_ImportFindingTypeEnumTypeWrapper): ...

IMPORT_FINDING_TYPE_NONE: ImportFindingType.ValueType  # 0
IMPORT_FINDING_TYPE_DELETED: ImportFindingType.ValueType  # 1
IMPORT_FINDING_TYPE_INVALID_JSON: ImportFindingType.ValueType  # 2
IMPORT_FINDING_TYPE_INVALID_PACKAGE: ImportFindingType.ValueType  # 3
IMPORT_FINDING_TYPE_INVALID_PURL: ImportFindingType.ValueType  # 4
IMPORT_FINDING_TYPE_INVALID_VERSION: ImportFindingType.ValueType  # 5
IMPORT_FINDING_TYPE_INVALID_COMMIT: ImportFindingType.ValueType  # 6
IMPORT_FINDING_TYPE_INVALID_RANGE: ImportFindingType.ValueType  # 7
IMPORT_FINDING_TYPE_INVALID_RECORD: ImportFindingType.ValueType  # 8
IMPORT_FINDING_TYPE_INVALID_ALIASES: ImportFindingType.ValueType  # 9
IMPORT_FINDING_TYPE_INVALID_UPSTREAM: ImportFindingType.ValueType  # 10
IMPORT_FINDING_TYPE_INVALID_RELATED: ImportFindingType.ValueType  # 11
IMPORT_FINDING_TYPE_BAD_ALIASED_CVE: ImportFindingType.ValueType  # 12
global___ImportFindingType = ImportFindingType

@typing.final
class ImportFinding(google.protobuf.message.Message):
    """An importfinding entry.
    The protobuf representation is *NOT* stable and only used for implementing
    the JSON based API.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUG_ID_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    FINDINGS_FIELD_NUMBER: builtins.int
    FIRST_SEEN_FIELD_NUMBER: builtins.int
    LAST_ATTEMPT_FIELD_NUMBER: builtins.int
    bug_id: builtins.str
    source: builtins.str
    @property
    def findings(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[global___ImportFindingType.ValueType]: ...
    @property
    def first_seen(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def last_attempt(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    def __init__(
        self,
        *,
        bug_id: builtins.str = ...,
        source: builtins.str = ...,
        findings: collections.abc.Iterable[global___ImportFindingType.ValueType] | None = ...,
        first_seen: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        last_attempt: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["first_seen", b"first_seen", "last_attempt", b"last_attempt"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bug_id", b"bug_id", "findings", b"findings", "first_seen", b"first_seen", "last_attempt", b"last_attempt", "source", b"source"]) -> None: ...

global___ImportFinding = ImportFinding



#!/usr/bin/env python3

import os
from html.parser import HTMLParser


class UnexpectedTagError(Exception):
  pass


class HTMLValidator(HTMLParser):
  def __init__(self) -> None:
    super().__init__()

    self.__tags: list[str] = []

  def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
    self.__tags.append(tag)

  def handle_endtag(self, tag: str) -> None:
    last_tag = self.__tags.pop()
    if last_tag != tag:
      raise UnexpectedTagError(last_tag)


def extract_schema_table() -> tuple[str, int]:
  raw_table = ''
  index = -1

  with open('docs/schema.md') as f:
    for i, line in enumerate(f.readlines()):
      if line == '<table>\n':
        index = i + 1
        raw_table += line
      if raw_table != '':
        raw_table += line
      if line == '</table>\n':
        break
  return raw_table, index


table, starting_line = extract_schema_table()

validator = HTMLValidator()

try:
  validator.feed(table)
except UnexpectedTagError as e:
  if 'CI' in os.environ:
    print(
      f'::error file=docs/schema.md,line={starting_line}::unexpected {e} tag in table - ensure that the tags are properly paired'
    )
  print(
    f'unexpected {e} tag in docs/schema.md databases table - ensure that the tags are properly paired'
  )
  exit(1)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Models tests."""

import datetime
import os
import unittest

from . import models

from . import bug
from . import gcs
from . import tests
from . import vulnerability_pb2

from google.cloud import ndb


class ModelsTest(unittest.TestCase):
  """Tests for ndb Model migrations."""

  def setUp(self):
    models.SourceRepository(
        id='test',
        name='test',
        db_prefix=['TEST-'],
    ).put()
    return super().setUp()

  def test_bug_post_put(self):
    """Test _post_put_hook for Bug to populate new datastore/gcs entities."""
    vuln_id = 'TEST-123'
    # Create a handmade populated Bug
    models.AliasGroup(
        bug_ids=sorted([vuln_id, 'CVE-123', 'OSV-123']),
        last_modified=datetime.datetime(2025, 3, 4, tzinfo=datetime.UTC)).put()
    models.UpstreamGroup(
        db_id=vuln_id,
        upstream_ids=['TEST-1', 'TEST-12'],
        last_modified=datetime.datetime(2025, 3, 5, tzinfo=datetime.UTC)).put()
    models.Bug(
        db_id=vuln_id,
        aliases=['CVE-123'],
        related=['TEST-234'],
        upstream_raw=['TEST-12'],
        summary='This is a vuln',
        severities=[
            models.Severity(type='CVSS_V2', score='AV:N/AC:L/Au:S/C:P/I:P/A:N')
        ],
        status=bug.BugStatus.PROCESSED,
        timestamp=datetime.datetime(2025, 1, 1, tzinfo=datetime.UTC),
        last_modified=datetime.datetime(2025, 3, 3, tzinfo=datetime.UTC),
        import_last_modified=datetime.datetime(2025, 2, 2, tzinfo=datetime.UTC),
        source_id=f'test:{vuln_id}.json',
        source_of_truth=models.SourceOfTruth.SOURCE_REPO,
        public=True,
        affected_packages=[
            models.AffectedPackage(
                package=models.Package(ecosystem='npm', name='testjs'),
                ranges=[
                    models.AffectedRange2(
                        type='SEMVER',
                        events=[
                            models.AffectedEvent(type='fixed', value='1.0.0'),
                            models.AffectedEvent(type='introduced', value='0'),
                        ]),
                    models.AffectedRange2(
                        type='SEMVER',
                        events=[
                            models.AffectedEvent(
                                type='last_affected', value='2.2.0'),
                            models.AffectedEvent(
                                type='introduced', value='2.0.0'),
                        ])
                ],
                versions=['0.1.0', '0.2.0', '0.3.0', '2.0.0', '2.1.0',
                          '2.2.0']),
            models.AffectedPackage(
                package=models.Package(
                    ecosystem='Ubuntu:24.04:LTS', name='test'),
                ranges=[
                    models.AffectedRange2(
                        type='ECOSYSTEM',
                        events=[
                            models.AffectedEvent(type='introduced', value='0'),
                            models.AffectedEvent(type='fixed', value='1.0.0-3'),
                        ])
                ],
                versions=['1.0.0-1', '1.0.0-2'],
                severities=[models.Severity(type='Ubuntu', score='Low')]),
            models.AffectedPackage(
                package=models.Package(ecosystem='Ubuntu:25.04', name='test'),
                ranges=[
                    models.AffectedRange2(
                        type='ECOSYSTEM',
                        events=[
                            models.AffectedEvent(type='introduced', value='0'),
                            models.AffectedEvent(type='fixed', value='1.0.0-3'),
                        ])
                ],
                versions=['1.0.0-1', '1.0.0-2'],
                severities=[models.Severity(type='Ubuntu', score='High')]),
            models.AffectedPackage(
                package=models.Package(ecosystem='', name=''),
                ranges=[
                    models.AffectedRange2(
                        type='GIT', repo_url='https://github.com/test/test')
                ],
                versions=['v1', 'v2']),
        ],
    ).put()
    put_bug = models.Bug.get_by_id(vuln_id)
    self.assertIsNotNone(put_bug)
    put_bug: models.Bug

    # Check if new db entities were created.
    vulnerability = models.Vulnerability.get_by_id(vuln_id)
    self.assertIsNotNone(vulnerability)
    vulnerability: models.Vulnerability
    self.assertEqual('test:TEST-123.json', vulnerability.source_id)
    self.assertEqual(
        datetime.datetime(2025, 3, 5, tzinfo=datetime.UTC),
        vulnerability.modified)
    self.assertFalse(vulnerability.is_withdrawn)
    self.assertEqual(
        datetime.datetime(2025, 2, 2, tzinfo=datetime.UTC),
        vulnerability.modified_raw)
    self.assertListEqual(['CVE-123'], vulnerability.alias_raw)
    self.assertListEqual(['TEST-234'], vulnerability.related_raw)
    self.assertListEqual(['TEST-12'], vulnerability.upstream_raw)

    listed_vuln = models.ListedVulnerability.get_by_id(vuln_id)
    self.assertIsNotNone(listed_vuln)
    listed_vuln: models.ListedVulnerability
    self.assertEqual(
        datetime.datetime(2025, 1, 1, tzinfo=datetime.UTC),
        listed_vuln.published)
    self.assertListEqual(['GIT', 'Ubuntu', 'npm'], listed_vuln.ecosystems)
    self.assertListEqual([
        'Ubuntu:24.04:LTS/test', 'Ubuntu:25.04/test', 'github.com/test/test',
        'npm/testjs'
    ], listed_vuln.packages)
    self.assertEqual('This is a vuln', listed_vuln.summary)
    self.assertTrue(listed_vuln.is_fixed)
    self.assertListEqual([
        models.Severity(type='CVSS_V2', score='AV:N/AC:L/Au:S/C:P/I:P/A:N'),
        models.Severity(type='Ubuntu', score='High'),
        models.Severity(type='Ubuntu', score='Low')
    ], listed_vuln.severities)
    self.assertListEqual(
        ['https://github.com/test/test', 'test', 'test-123', 'testjs'],
        listed_vuln.autocomplete_tags)
    # search_indices should include all the original search indices,
    # plus the transitive alias & upstream ids
    search_indices = sorted(put_bug.search_indices +
                            ['osv-123', 'osv', 'test-1', '1'])
    self.assertListEqual(search_indices, listed_vuln.search_indices)

    affected: list[models.AffectedVersions] = models.AffectedVersions.query(
        models.AffectedVersions.vuln_id == vuln_id).fetch()
    affected.sort(key=lambda x: x.sort_key())
    want = [
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='GIT',
            name='github.com/test/test',
            versions=['v1', 'v2']),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu',
            name='test',
            events=[
                models.AffectedEvent(type='introduced', value='0'),
                models.AffectedEvent(type='fixed', value='1.0.0-3')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu',
            name='test',
            versions=['1.0.0-1', '1.0.0-2']),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:24.04',
            name='test',
            events=[
                models.AffectedEvent(type='introduced', value='0'),
                models.AffectedEvent(type='fixed', value='1.0.0-3')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:24.04',
            name='test',
            versions=['1.0.0-1', '1.0.0-2']),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:24.04:LTS',
            name='test',
            events=[
                models.AffectedEvent(type='introduced', value='0'),
                models.AffectedEvent(type='fixed', value='1.0.0-3')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:24.04:LTS',
            name='test',
            versions=['1.0.0-1', '1.0.0-2']),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:25.04',
            name='test',
            events=[
                models.AffectedEvent(type='introduced', value='0'),
                models.AffectedEvent(type='fixed', value='1.0.0-3')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='Ubuntu:25.04',
            name='test',
            versions=['1.0.0-1', '1.0.0-2']),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='npm',
            name='testjs',
            events=[
                models.AffectedEvent(type='introduced', value='0'),
                models.AffectedEvent(type='fixed', value='1.0.0')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='npm',
            name='testjs',
            events=[
                models.AffectedEvent(type='introduced', value='2.0.0'),
                models.AffectedEvent(type='last_affected', value='2.2.0')
            ]),
        models.AffectedVersions(
            vuln_id=vuln_id,
            ecosystem='npm',
            name='testjs',
            versions=['0.1.0', '0.2.0', '0.3.0', '2.0.0', '2.1.0', '2.2.0']),
    ]
    self.assertListEqual([a.to_dict() for a in want],
                         [a.to_dict() for a in affected])

    # Check the records written to the 'bucket' (which is mocked) are expected.
    vuln_pb = put_bug.to_vulnerability(True, True, True)

    bucket = gcs.get_osv_bucket()
    blob = bucket.get_blob(os.path.join(gcs.VULN_PB_PATH, f'{vuln_id}.pb'))
    self.assertIsNotNone(blob)
    self.assertEqual(blob.custom_time,
                     datetime.datetime(2025, 3, 5, tzinfo=datetime.UTC))
    got_pb = vulnerability_pb2.Vulnerability().FromString(
        blob.download_as_bytes())
    self.assertEqual(got_pb, vuln_pb)

  def test_bug_withdraw(self):
    """Test if withdrawing a Bug correctly removes unneeded indices."""
    # First put the bug un-withdrawn
    vuln_id = 'TEST-999'
    models.Bug(
        db_id=vuln_id,
        status=bug.BugStatus.PROCESSED,
        timestamp=datetime.datetime(2025, 1, 1, tzinfo=datetime.UTC),
        last_modified=datetime.datetime(2025, 3, 3, tzinfo=datetime.UTC),
        import_last_modified=datetime.datetime(2025, 2, 2, tzinfo=datetime.UTC),
        source_id=f'test:{vuln_id}.json',
        source_of_truth=models.SourceOfTruth.SOURCE_REPO,
        public=True,
        affected_packages=[
            models.AffectedPackage(
                package=models.Package(ecosystem='PyPI', name='testpy'),
                ranges=[
                    models.AffectedRange2(
                        type='ECOSYSTEM',
                        events=[
                            models.AffectedEvent(type='introduced', value='0'),
                            models.AffectedEvent(type='fixed', value='1.0'),
                        ])
                ],
                versions=['0.1', '0.2'],
            ),
        ],
    ).put()
    put_bug = models.Bug.get_by_id(vuln_id)
    self.assertIsNotNone(put_bug)
    put_bug: models.Bug

    vulnerability = models.Vulnerability.get_by_id(vuln_id)
    self.assertIsNotNone(vulnerability)
    vulnerability: models.Vulnerability
    self.assertFalse(vulnerability.is_withdrawn)
    listed_vuln = models.ListedVulnerability.get_by_id(vuln_id)
    self.assertIsNotNone(listed_vuln)
    affected = models.AffectedVersions.query(
        models.AffectedVersions.vuln_id == vuln_id).fetch()
    self.assertEqual(2, len(affected))

    bucket = gcs.get_osv_bucket()
    blob = bucket.get_blob(os.path.join(gcs.VULN_PB_PATH, f'{vuln_id}.pb'))
    self.assertIsNotNone(blob)
    self.assertEqual(blob.custom_time,
                     datetime.datetime(2025, 3, 3, tzinfo=datetime.UTC))

    # Now withdraw the Bug
    put_bug.withdrawn = datetime.datetime(2025, 4, 4, tzinfo=datetime.UTC)
    put_bug.last_modified = datetime.datetime(2025, 4, 4, tzinfo=datetime.UTC)
    put_bug.put()

    # Vulnerability exists, but is withdrawn
    vulnerability = models.Vulnerability.get_by_id(vuln_id)
    self.assertIsNotNone(vulnerability)
    vulnerability: models.Vulnerability
    self.assertTrue(vulnerability.is_withdrawn)
    # ListedVulnerability and AffectedVersions have been removed
    listed_vuln = models.ListedVulnerability.get_by_id(vuln_id)
    self.assertIsNone(listed_vuln)
    affected = models.AffectedVersions.query(
        models.AffectedVersions.vuln_id == vuln_id).fetch()
    self.assertEqual(0, len(affected))
    # Blobs still exist, and were re-written
    bucket = gcs.get_osv_bucket()
    blob = bucket.get_blob(os.path.join(gcs.VULN_PB_PATH, f'{vuln_id}.pb'))
    self.assertIsNotNone(blob)
    self.assertEqual(blob.custom_time,
                     datetime.datetime(2025, 4, 4, tzinfo=datetime.UTC))

  def test_oss_fuzz_private(self):
    """Test that non-public Bugs from OSS-Fuzz are not indexed."""
    vuln_id = 'TEST-OSSFUZZ'
    models.Bug(
        db_id=vuln_id,
        status=bug.BugStatus.UNPROCESSED,
        public=False,
        timestamp=datetime.datetime(2025, 1, 1, tzinfo=datetime.UTC),
        last_modified=datetime.datetime(2025, 3, 3, tzinfo=datetime.UTC),
        import_last_modified=datetime.datetime(2025, 2, 2, tzinfo=datetime.UTC),
        source_id=f'test:{vuln_id}.json',
        source_of_truth=models.SourceOfTruth.SOURCE_REPO,
        affected_packages=[
            models.AffectedPackage(
                package=models.Package(ecosystem='PyPI', name='testpy'),
                ranges=[
                    models.AffectedRange2(
                        type='ECOSYSTEM',
                        events=[
                            models.AffectedEvent(type='introduced', value='0'),
                            models.AffectedEvent(type='fixed', value='1.0'),
                        ])
                ],
                versions=['0.1', '0.2'],
            ),
        ],
    ).put()

    vulnerability = models.Vulnerability.get_by_id(vuln_id)
    self.assertIsNone(vulnerability)
    listed_vuln = models.ListedVulnerability.get_by_id(vuln_id)
    self.assertIsNone(listed_vuln)
    affected = models.AffectedVersions.query(
        models.AffectedVersions.vuln_id == vuln_id).fetch()
    self.assertEqual(0, len(affected))
    bucket = gcs.get_osv_bucket()
    blob = bucket.get_blob(os.path.join(gcs.VULN_PB_PATH, f'{vuln_id}.pb'))
    self.assertIsNone(blob)

  def test_normalize_repo(self):
    """Test normalize_repo_package function."""
    test_cases = [
        # protocol normalization
        ('http://git.musl-libc.org/git/musl', 'git.musl-libc.org/git/musl'),
        ('https://git.musl-libc.org/git/musl', 'git.musl-libc.org/git/musl'),
        ('git://git.musl-libc.org/git/musl', 'git.musl-libc.org/git/musl'),

        # github examples
        ('http://github.com/user/repo', 'github.com/user/repo'),
        ('https://github.com/user/repo', 'github.com/user/repo'),
        ('git://github.com/user/repo', 'github.com/user/repo'),

        # trailing slash
        ('https://github.com/user/repo/', 'github.com/user/repo'),
        ('http://git.example.com/path/', 'git.example.com/path'),

        # .git suffix removed
        ('https://github.com/user/repo.git', 'github.com/user/repo'),
        ('http://git.example.com/repo.git', 'git.example.com/repo'),

        # edge cases
        ('', ''),
        ('invalid-url', 'invalid-url'),
        ('http://', ''),
        ('https://hostname', 'hostname'),
    ]

    for repo_url, expected in test_cases:
      with self.subTest(repo_url=repo_url):
        self.assertEqual(expected, models.normalize_repo_package(repo_url))


def setUpModule():
  """Set up the test module."""
  # Start the emulator BEFORE creating the ndb client
  unittest.enterModuleContext(tests.datastore_emulator())
  unittest.enterModuleContext(ndb.Client().context(cache_policy=False))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CRAN helpers."""

import requests
import packaging_legacy.version

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError


class CRAN(EnumerableEcosystem):
  """CRAN ecosystem helpers."""

  # Use the Posit Public Package Manager API to pull both the current
  # and archived versions for a specific package since CRAN doesn't
  # natively support this functionality.
  _API_PACKAGE_URL_POSIT_CRAN = 'https://packagemanager.posit.co/__api__/' + \
    'repos/2/packages/{package}'

  def sort_key(self, version):
    """Sort key."""
    # Some documentation on CRAN versioning and the R numeric_version method:
    # https://cran.r-project.org/doc/manuals/R-exts.html#The-DESCRIPTION-file
    # https://stat.ethz.ch/R-manual/R-devel/library/base/html/numeric_version.html
    # The packaging.version appears to work for the typical X.Y.Z and
    # X.Y-Z cases
    version = version.replace("-", ".")
    # version.parse() handles invalid versions by returning LegacyVersion()
    return packaging_legacy.version.parse(version)

  def _enumerate_versions(self,
                          url,
                          package,
                          introduced,
                          fixed=None,
                          last_affected=None,
                          limits=None):
    """Helper method to enumerate versions from a specific URL."""
    response = requests.get(url.format(package=package), timeout=config.timeout)
    if response.status_code == 404:
      return None

    if response.status_code != 200:
      raise RuntimeError(
          f'Failed to get R versions for {package} with: {response.text}')

    response = response.json()
    versions = []
    if 'version' in response:
      versions = [response['version']]
    if 'archived' in response:
      versions += [archived['version'] for archived in response['archived']]

    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    versions = self._enumerate_versions(self._API_PACKAGE_URL_POSIT_CRAN,
                                        package, introduced, fixed,
                                        last_affected, limits)

    if versions is None:
      raise EnumerateError(f'Package {package} not found')

    return versions



# Copyright 2025 Google LLC
# Copyright 2023 Fraser Tweedale
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Haskell ecosystem helpers.

Contact the Haskell Security Response Team <security-advisories@haskell.org>
if something is broken and you need help to fix it.

"""

import requests

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .semver_ecosystem_helper import SemverLike


class Hackage(EnumerableEcosystem):
  """Hackage (Haskell package index) ecosystem."""

  _API_PACKAGE_URL = 'https://hackage.haskell.org/package/{package}.json'

  def sort_key(self, version):
    """Sort key.

    The Haskell package version data type is defined at
    https://hackage.haskell.org/package/Cabal-syntax/docs/Distribution-Types-Version.html

    """
    # If version is not valid, it is most likely an invalid input version
    # then sort it to the last/largest element
    try:
      return [int(x) for x in version.split('.')]
    except ValueError:
      return [999999]

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    response = requests.get(
        self._API_PACKAGE_URL.format(package=package), timeout=config.timeout)
    if response.status_code == 404:
      raise EnumerateError(f'Package {package} not found')
    if response.status_code != 200:
      raise RuntimeError(
          f'Failed to get Hackage versions for {package} with: {response.text}')

    response = response.json()
    versions = list(response.keys())

    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)


class GHC(EnumerableEcosystem, SemverLike):
  """Glasgow Haskell Compiler (GHC) ecosystem."""

  _API_PACKAGE_URL = ('https://gitlab.haskell.org'
                      '/api/v4/projects/3561/repository/tags?per_page=100')
  # 100 is the maximum per_page size according to GitLab docs:
  # https://docs.gitlab.com/ee/api/rest/index.html#offset-based-pagination
  """
  Historical versions do not have tags in the Git repo, so we hardcode the
  list.  See https://github.com/google/osv.dev/pull/1463 for discussion.
  """
  HISTORICAL_VERSIONS = [
    '0.29',
    '2.10',
    '3.02', '3.03',
    '4.02', '4.04', '4.06', '4.08', '4.08.1', '4.08.2',
    '5.00', '5.00.1', '5.00.2', '5.02', '5.02.1', '5.02.2', '5.02.3',
    '5.04', '5.04.1', '5.04.2', '5.04.3',
    '6.0', '6.0.1',
    '6.2', '6.2.1', '6.2.2',
    '6.4', '6.4.1', '6.4.2', '6.4.3',
    '6.6', '6.6.1',
    '6.8.1', '6.8.1', '6.8.3',
    '6.10.1', '6.10.2-rc1', '6.10.2', '6.10.3', '6.10.4-rc1', '6.10.4',
    '6.12.1-rc1', '6.12.1', '6.12.2-rc1', '6.12.2', '6.12.3-rc1', '6.12.3',
    '7.0.1-rc1', '7.0.1-rc2', '7.0.1', '7.0.2-rc1', '7.0.2-rc2', '7.0.2',
    '7.0.3', '7.0.4-rc1', '7.0.4',
  ]  # yapf: disable


  @classmethod
  def tag_to_version(cls, tag: str) -> str | None:
    """Convert a tag to a release version, or return None if invalid.

    GHC release tags follow the scheme:

    - ghc-<major>.<minor>.<patch>-alpha<n>
    - ghc-<major>.<minor>.<patch>-rc<n>
    - ghc-<major>.<minor>.<patch>-release

    """
    parts = tag.split('-')
    if len(parts) == 3 and parts[0] == 'ghc' \
        and cls.is_major_minor_patch(parts[1]):
      if parts[2].startswith('alpha') or parts[2].startswith('rc'):
        return '-'.join(parts[1:3])
      if parts[2] == 'release':
        return parts[1]
    return None

  @staticmethod
  def is_major_minor_patch(s: str) -> bool:
    """Check that string matches ``<int>.<int>.<int>``."""
    parts = s.split('.')
    return len(parts) == 3 and all(x.isdigit() for x in parts)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions.

    Different components of GHC are part of the same software release.
    So we ignore the package (component) name.

    """

    # Versions come from tags from the GitLab API, which are paginated.
    # https://docs.gitlab.com/ee/api/rest/index.html#pagination-link-header
    url = self._API_PACKAGE_URL
    versions = self.HISTORICAL_VERSIONS.copy()
    while url is not None:
      response = requests.get(url, timeout=config.timeout)
      if response.status_code == 404:
        raise EnumerateError(f'GHC tag list not found at {url}')
      if response.status_code != 200:
        raise RuntimeError(
            f'Failed to get GHC versions from: {url} with: {response.text}')

      versions.extend(self.tag_to_version(x['name']) for x in response.json())

      if 'next' not in response.links:
        break
      url = response.links['next'].get('url')

    versions = [v for v in versions if v is not None]

    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PURL conversion utilities."""

from collections import namedtuple
from urllib.parse import quote

from packageurl import PackageURL

ParsedPURL = namedtuple('ParsedPURL', ['ecosystem', 'package', 'version'])
EcosystemPURL = namedtuple('EcosystemPURL', ['type', 'namespace'])

# PURL spec:
# https://github.com/package-url/purl-spec/blob/master/PURL-TYPES.rst
ECOSYSTEM_PURL_DATA = {
    'AlmaLinux':
        EcosystemPURL('rpm', 'almalinux'),
    'Alpaquita':
        EcosystemPURL('apk', 'alpaquita'),
    'Alpine':
        EcosystemPURL('apk', 'alpine'),
    # Android
    'BellSoft Hardened Containers':
        EcosystemPURL('apk', 'bellsoft-hardened-containers'),
    # Bioconductor
    'Bitnami':
        EcosystemPURL('bitnami', None),
    'Chainguard':
        EcosystemPURL('apk', 'chainguard'),
    'ConanCenter':
        EcosystemPURL('conan', None),
    'CRAN':
        EcosystemPURL('cran', None),
    'crates.io':
        EcosystemPURL('cargo', None),
    'Echo':
        EcosystemPURL('deb', 'echo'),
    'Debian':
        EcosystemPURL('deb', 'debian'),
    # GHC
    # GIT
    # GitHub Actions
    'Go':
        EcosystemPURL('golang', None),
    'Hackage':
        EcosystemPURL('hackage', None),
    'Hex':
        EcosystemPURL('hex', None),
    # Linux
    'Mageia':
        EcosystemPURL('rpm', 'mageia'),
    'Maven':
        EcosystemPURL('maven', None),
    'MinimOS':
        EcosystemPURL('apk', 'minimos'),
    'npm':
        EcosystemPURL('npm', None),
    'NuGet':
        EcosystemPURL('nuget', None),
    'openEuler':
        EcosystemPURL('rpm', 'openeuler'),
    'openSUSE':
        EcosystemPURL('rpm', 'opensuse'),
    'OSS-Fuzz':
        EcosystemPURL('generic', None),
    'Packagist':
        EcosystemPURL('composer', None),
    # Photon OS
    'Pub':
        EcosystemPURL('pub', None),
    'PyPI':
        EcosystemPURL('pypi', None),
    'Red Hat':
        EcosystemPURL('rpm', 'redhat'),
    'Rocky Linux':
        EcosystemPURL('rpm', 'rocky-linux'),
    'RubyGems':
        EcosystemPURL('gem', None),
    'SUSE':
        EcosystemPURL('rpm', 'suse'),
    'SwiftURL':
        EcosystemPURL('swift', None),
    'Ubuntu':
        EcosystemPURL('deb', 'ubuntu'),
    'Wolfi':
        EcosystemPURL('apk', 'wolfi'),
}

# Create the reverse lookup hash map
PURL_ECOSYSTEM_MAP = {
    purl_data: ecosystem
    for ecosystem, purl_data in ECOSYSTEM_PURL_DATA.items()
}


def _url_encode(package_name):
  """URL encode a PURL `namespace/name` or `name`."""
  parts = package_name.split('/')
  return '/'.join(quote(p) for p in parts)


def package_to_purl(ecosystem: str, package_name: str) -> str | None:
  """Convert a ecosystem and package name to PURL."""
  purl_data = ECOSYSTEM_PURL_DATA.get(ecosystem)
  if not purl_data:
    return None

  purl_type, purl_namespace = purl_data
  if purl_namespace:
    purl_ecosystem = f'{purl_type}/{purl_namespace}'
  else:
    purl_ecosystem = purl_type

  suffix = ''

  if purl_type == 'maven':
    # PURLs use / to separate the group ID and the artifact ID.
    package_name = package_name.replace(':', '/', 1)

  if purl_type == 'deb' and ecosystem == 'Debian':
    suffix = '?arch=source'

  if purl_type == 'apk' and ecosystem in ('Alpine', 'Alpaquita',
                                          'BellSoft Hardened Containers'):
    suffix = '?arch=source'

  return f'pkg:{purl_ecosystem}/{_url_encode(package_name)}{suffix}'


def parse_purl(purl_str: str) -> ParsedPURL | None:
  """Parses a PURL string and extracts
  ecosystem, package, and version information.

  Args:
    purl_str: The Package URL string to parse.

  Returns:
    A tuple containing the ecosystem, package,
    and version, or None if parsing fails.
  """

  purl = PackageURL.from_string(
      purl_str)  # May raise ValueError for invalid PURL strings

  package = purl.name
  version = purl.version

  # Find a matching ecosystem using both type and namespace.
  ecosystem = PURL_ECOSYSTEM_MAP.get(EcosystemPURL(purl.type, purl.namespace))
  if ecosystem:
    return ParsedPURL(ecosystem, package, version)

  # If no match is found, try again using only the type.
  # Some ecosystems may use the namespace to represent additional
  # information (like vendors) and the namespace might be optional.
  ecosystem = PURL_ECOSYSTEM_MAP.get(EcosystemPURL(purl.type, None))
  if not ecosystem:
    return None

  # For ecosystems with optional namespaces, the namespace might need to be
  # included as part of the package name.
  if purl.namespace:
    if purl.type == 'golang':
      package = purl.namespace + '/' + purl.name
      if purl.subpath:
        package = package + '/' + purl.subpath
    elif purl.type in ('composer', 'hex', 'npm', 'swift'):
      package = purl.namespace + '/' + purl.name
    elif purl.type == 'maven':
      package = purl.namespace + ':' + purl.name
    else:
      # Handle the case where the ecosystem shouldn't have a namespace.
      return None
  else:
    # Handle the case where the namespace is not supported.
    return None

  return ParsedPURL(ecosystem, package, version)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SemVer indexer."""

import re

import semver

_PAD_WIDTH = 8
_FAKE_PRE_WIDTH = 16


def _strip_leading_v(version):
  """Strip leading v from the version, if any."""
  # Versions starting with "v" aren't valid SemVer, but we handle them just in
  # case.
  if version.startswith('v'):
    return version[1:]

  return version


def _remove_leading_zero(component):
  """Remove leading zeros from a component."""
  if component[0] == '.':
    return '.' + str(int(component[1:]))

  return str(int(component))


def _coerce_suffix(suffix):
  """Coerce a potentially invalid semver suffix into a valid semver suffix.

  Removes leading zeros from the pre-release suffix and modifies empty
  components to allow earlier SemVer specs (specifically 2.0.0-rc.1) to be
  parsed.

  Empty components are replaced with '-' (i.e 1.0.0-a..0 -> 1.0.0-a.-.0) which
  mostly preserves ordering."""

  if not suffix:
    return suffix

  suffix_pattern = re.compile(r'^(-[^+]*)?(\+.*)?(.*)$')
  match = suffix_pattern.match(suffix)

  pre = ''
  if match.group(1):
    pre_components = []
    for component in match.group(1)[1:].split('.'):
      if not component:
        pre_components.append('-')
      elif component.isdigit():
        pre_components.append(_remove_leading_zero(component))
      else:
        pre_components.append(component)
    pre = '-' + '.'.join(pre_components)

  build = ''
  if match.group(2):
    build_components = []
    for component in match.group(2)[1:].split('.'):
      if not component:
        build_components.append('-')
      else:
        build_components.append(component)
    build = '+' + '.'.join(build_components)

  return pre + build + match.group(3)


def coerce(version: str):
  """Coerce a potentially invalid semver into valid semver."""
  version = _strip_leading_v(version)
  version_pattern = re.compile(r'^(\d+)(\.\d+)?(\.\d+)?(.*)$')
  match = version_pattern.match(version)
  if not match:
    return version

  return (_remove_leading_zero(match.group(1)) +
          _remove_leading_zero(match.group(2) or '.0') +
          _remove_leading_zero(match.group(3) or '.0') +
          _coerce_suffix(match.group(4)))


def is_valid(version):
  """Returns whether or not the version is a valid semver."""
  return semver.Version.is_valid(_strip_leading_v(version))


def parse(version: str):
  """Parse a SemVer."""
  return semver.Version.parse(coerce(version))


def normalize(version):
  """Normalize semver version for indexing (to allow for lexical
  sorting/filtering)."""
  version = parse(version)

  # Precedence rules: https://semver.org/#spec-item-11

  # 1. Precedence MUST be calculated by separating the version into major,
  # minor, patch and pre-release identifiers in that order (Build metadata does
  # not figure into precedence).
  #
  # Normalization: Per spec build metadata is ignored.

  # 2. Precedence is determined by the first difference when comparing each of
  # these identifiers from left to right as follows: Major, minor, and patch
  # versions are always compared numerically.
  #
  # Normalization: Pad the components with '0' to allow for lexical ordering of
  # numbers.
  core_parts = '{}.{}.{}'.format(
      str(version.major).rjust(_PAD_WIDTH, '0'),
      str(version.minor).rjust(_PAD_WIDTH, '0'),
      str(version.patch).rjust(_PAD_WIDTH, '0'))

  # 3. When major, minor, and patch are equal, a pre-release version has lower
  # precedence than a normal version:
  #
  # Normalization: Attach a very long fake prerelease version that is most
  # likely to come after any real prelease version.
  if not version.prerelease:
    pre = 'z' * _FAKE_PRE_WIDTH
    return f'{core_parts}-{pre}'

  pre = normalize_prerelease(version.prerelease)
  return f'{core_parts}-{pre}'


def normalize_prerelease(prerelease):
  """Normalize semver pre-release version suffix for indexing (to allow for
  lexical sorting/filtering)."""
  # 4. Precedence for two pre-release versions with the same major, minor, and
  # patch version MUST be determined by comparing each dot separated identifier
  # from left to right until a difference is found as follows:
  #
  # Normalization: Pad the components.
  pre_components = []
  for component in prerelease.split('.'):
    # 3. Numeric identifiers always have lower precedence than non-numeric
    # identifiers.
    #
    # Normalization: Pad numeric components with '0', and prefix alphanumeric
    # with a single '1' (to ensure they always come after).
    if component.isdigit():
      # 1. Identifiers consisting of only digits are compared numerically.
      pre_components.append(component.rjust(_PAD_WIDTH, '0'))
    else:
      # 2. Identifiers with letters or hyphens are compared lexically in ASCII
      # sort order.
      pre_components.append('1' + component)

  # 4. A larger set of pre-release fields has a higher precedence than a smaller
  # set, if all of the preceding identifiers are equal.
  #
  # Consistent with lexical sorting after normalization.
  return '.'.join(pre_components)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyPI ecosystem helper."""

import packaging_legacy.version
import requests

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError


class PyPI(EnumerableEcosystem):
  """PyPI ecosystem helpers."""

  _API_PACKAGE_URL = 'https://pypi.org/pypi/{package}/json'

  def sort_key(self, version):
    """Sort key."""
    # version.parse() handles invalid versions by returning LegacyVersion()
    return packaging_legacy.version.parse(version)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    response = requests.get(
        self._API_PACKAGE_URL.format(package=package), timeout=config.timeout)

    if response.status_code == 404:
      raise EnumerateError(f'Package {package} not found')
    if response.status_code != 200:
      raise RuntimeError(
          f'Failed to get PyPI versions for {package} with: {response.text}')

    response = response.json()
    versions = list(response['releases'].keys())
    self.sort_versions(versions)

    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helper tests."""

import unittest
from .. import ecosystems


class EcosystemTest(unittest.TestCase):
  """Ecosystem helper tests."""

  def test_add_matching_ecosystems(self):
    """Test sort key"""
    # Test Ubuntu
    ubuntu_ecosystem = {
        'Ubuntu', 'Ubuntu:20.04:LTS', 'Ubuntu:22.04:LTS', 'Ubuntu:24.04:LTS',
        'Ubuntu:24.10', 'Ubuntu:Pro:14.04:LTS', 'Ubuntu:Pro:16.04:LTS',
        'Ubuntu:Pro:18.04:LTS'
    }
    actual_output = list(ecosystems.add_matching_ecosystems(ubuntu_ecosystem))
    expected_output = [
        'Ubuntu', 'Ubuntu:14.04', 'Ubuntu:16.04', 'Ubuntu:18.04',
        'Ubuntu:20.04', 'Ubuntu:20.04:LTS', 'Ubuntu:22.04', 'Ubuntu:22.04:LTS',
        'Ubuntu:24.04', 'Ubuntu:24.04:LTS', 'Ubuntu:24.10',
        'Ubuntu:Pro:14.04:LTS', 'Ubuntu:Pro:16.04:LTS', 'Ubuntu:Pro:18.04:LTS'
    ]
    actual_output.sort()
    self.assertEqual(list(actual_output), expected_output)

    #Test Debian (it should be no change)
    debian_ecosystem = {'Debian', 'Debian:11', 'Debian:12', 'Debian:13'}
    actual_output = list(ecosystems.add_matching_ecosystems(debian_ecosystem))
    expected_output = ['Debian', 'Debian:11', 'Debian:12', 'Debian:13']
    actual_output.sort()
    self.assertEqual(list(actual_output), expected_output)

  def test_maybe_normalize_package_names(self):
    """Test normalize package name"""
    package_name = 'Flask'
    ecosystem = 'PyPI'
    expected = 'flask'

    actual = ecosystems.maybe_normalize_package_names(package_name, ecosystem)
    self.assertEqual(actual, expected)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RPM / Red Hat Linux ecosystem helper tests."""

import unittest

from . import redhat

from .. import ecosystems


class RPMEcosystemTest(unittest.TestCase):
  """RPM ecosystem helper tests."""

  def test_rpm(self):
    """Test sort_key"""
    ecosystem = redhat.RPM()
    # Red Hat
    self.assertGreater(
        ecosystem.sort_key('0:0.2.6-20.module+el8.9.0+1420+91577025'),
        ecosystem.sort_key('0:0.0.99.4-5.module+el8.9.0+1445+07728297'))
    self.assertGreater(
        ecosystem.sort_key('0:0.2.6-20.module+el8.9.0+1420+91577025'),
        ecosystem.sort_key('0'))
    self.assertGreater(
        ecosystem.sort_key('2:1.14.3-2.module+el8.10.0+1815+5fe7415e'),
        ecosystem.sort_key('2:1.10.3-1.module+el8.10.0+1815+5fe7415e'))
    self.assertLess(ecosystem.sort_key('invalid'), ecosystem.sort_key('0'))

    # AlmaLinux
    self.assertGreater(
        ecosystem.sort_key("9.27-15.el8_10"),
        ecosystem.sort_key("9.27-13.el8_10"))
    self.assertGreater(
        ecosystem.sort_key("9.27-15.el8_10"), ecosystem.sort_key("0"))
    self.assertGreater(
        ecosystem.sort_key("3:2.1.10-1.module_el8.10.0+3858+6ad51f9f"),
        ecosystem.sort_key("3:2.1.10-1.module_el8.10.0+3845+87b84552"))
    self.assertLess(
        ecosystem.sort_key("20230404-117.git2e92a49f.el8_8.alma.1"),
        ecosystem.sort_key("20240111-121.gitb3132c18.el8"))
    self.assertEqual(
        ecosystem.sort_key("20240111-121.gitb3132c18.el8"),
        ecosystem.sort_key("20240111-121.gitb3132c18.el8"))

    # Mageia
    self.assertGreater(
        ecosystem.sort_key('3.2.7-1.2.mga9'),
        ecosystem.sort_key('3.2.7-1.mga9'))
    self.assertGreater(
        ecosystem.sort_key('3.2.7-1.2.mga9'), ecosystem.sort_key('0'))
    self.assertLess(ecosystem.sort_key('invalid'), ecosystem.sort_key('0'))
    self.assertGreater(
        ecosystem.sort_key('1:1.8.11-1.mga9'),
        ecosystem.sort_key('0:1.9.1-2.mga9'))

    # openEuler
    self.assertGreater(
        ecosystem.sort_key("1.2.3-1.oe2203"),
        ecosystem.sort_key("1.2.2-1.oe2203"))
    self.assertGreater(
        ecosystem.sort_key("2.0.0-1.oe2203"), ecosystem.sort_key("0"))
    self.assertGreater(
        ecosystem.sort_key("1.2.3-2.oe2203"),
        ecosystem.sort_key("1.2.3-1.oe2203"))
    self.assertLess(
        ecosystem.sort_key("1.2.2-1.oe2203"),
        ecosystem.sort_key("1.2.3-1.oe2203"))
    self.assertEqual(
        ecosystem.sort_key("1.2.3-1.oe2203"),
        ecosystem.sort_key("1.2.3-1.oe2203"))

    # openSUSE
    self.assertGreater(
        ecosystem.sort_key("4.2-lp151.4.3.1"),
        ecosystem.sort_key("1.5.1-lp151.4.3.1"))
    self.assertGreater(
        ecosystem.sort_key("4.9.6-bp152.2.3.1"), ecosystem.sort_key("0"))
    self.assertGreater(
        ecosystem.sort_key("6.2.8-bp156.2.3.1"),
        ecosystem.sort_key("6.2.8-bp156"))
    self.assertLess(
        ecosystem.sort_key("0.4.6-15.8"), ecosystem.sort_key("1.4.6-15.8"))
    self.assertEqual(
        ecosystem.sort_key("6.2.8-bp156.2.3.1"),
        ecosystem.sort_key("6.2.8-bp156.2.3.1"))

    # SUSE
    self.assertGreater(
        ecosystem.sort_key("2.38.5-150400.4.34.2"),
        ecosystem.sort_key("2.37.5-150400.4.34.2"))
    self.assertGreater(
        ecosystem.sort_key("2.0.8-4.8.2"), ecosystem.sort_key("0"))
    self.assertGreater(
        ecosystem.sort_key("2.0.8_k4.12.14_10.118-4.8.2"),
        ecosystem.sort_key("2.0.8-4.8.2"))
    self.assertLess(
        ecosystem.sort_key("1.86-150100.7.23.11"),
        ecosystem.sort_key("2.86-150100.7.23.1"))
    self.assertEqual(
        ecosystem.sort_key("2.0.8-4.8.2"), ecosystem.sort_key("2.0.8-4.8.2"))

    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('1.10.2-1.oe2203'),
        ecosystem.sort_key('1.2.2-1.oe2203'))
    self.assertLessEqual(
        ecosystem.sort_key('1.2.2-1.oe2203'),
        ecosystem.sort_key('1.10.2-1.oe2203'))

  def test_rpm_ecosystems(self):
    """Test RPM-based ecosystems return an RPM ecosystem."""
    ecos = [
        'Red Hat',
        'AlmaLinux',
        'Mageia',
        'openEuler',
        'openSUSE',
        'Rocky Linux',
        'SUSE',
    ]
    for ecosystem_name in ecos:
      ecosystem = ecosystems.get(ecosystem_name)
      self.assertIsInstance(ecosystem, redhat.RPM)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""NuGet ecosystem helper."""

import functools
import re
import requests

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .. import semver_index

# This relies on a strict SemVer implementation.
# Differences from SemVer are described at
# https://docs.microsoft.com/en-us/nuget/concepts/package-versioning
#  - Optional 4th component (x.y.z.R).
#  - Prerelease components are compared case insensitively.
#  - Non-major version segments are optional. e.g. "1" is a valid version
#    number.


def _extract_revision(str_version):
  """Extract revision (4th component) from version number (if any)."""
  # e.g. '1.0.0.0-prerelease'
  pattern = re.compile(r'^(\d+)(\.\d+)(\.\d+)(\.\d+)(.*)')
  match = pattern.match(str_version)
  if not match:
    return str_version, 0

  return (''.join(
      (match.group(1), match.group(2), match.group(3), match.group(5))),
          int(match.group(4)[1:]))


@functools.total_ordering
class Version:
  """NuGet version."""

  def __init__(self, base_semver, revision):
    self._base_semver = base_semver
    if self._base_semver.prerelease:
      self._base_semver = self._base_semver.replace(
          prerelease=base_semver.prerelease.lower())
    self._revision = revision

  def __eq__(self, other):
    return (self._base_semver == other._base_semver and
            self._revision == other._revision)

  def __lt__(self, other):
    if (self._base_semver.replace(prerelease='') == other._base_semver.replace(
        prerelease='')):
      # If the first three components are the same, compare the revision.
      if self._revision != other._revision:
        return self._revision < other._revision

    # Revision is the same, so ignore it for comparison purposes.
    return self._base_semver < other._base_semver

  @classmethod
  def from_string(cls, str_version):
    str_version = semver_index.coerce(str_version)
    str_version, revision = _extract_revision(str_version)
    try:
      return Version(semver_index.parse(str_version), revision)
    except ValueError:
      # If a user gives us an unparsable semver version,
      # treat it as a very large version so as to not match anything.
      return Version(semver_index.parse('999999'), 999999)


class NuGet(EnumerableEcosystem):
  """NuGet ecosystem."""

  _API_PACKAGE_URL = ('https://api.nuget.org/v3/registration5-semver1/'
                      '{package}/index.json')

  def sort_key(self, version):
    """Sort key."""
    return Version.from_string(version)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    url = self._API_PACKAGE_URL.format(package=package.lower())
    response = requests.get(url, timeout=config.timeout)
    if response.status_code == 404:
      raise EnumerateError(f'Package {package} not found')
    if response.status_code != 200:
      raise RuntimeError(
          f'Failed to get NuGet versions for {package} with: {response.text}')

    response = response.json()

    versions = []
    for page in response['items']:
      if 'items' in page:
        items = page['items']
      else:
        items_response = requests.get(page['@id'], timeout=config.timeout)
        if items_response.status_code != 200:
          raise RuntimeError(
              f'Failed to get NuGet versions page for {package} with: '
              f'{items_response.text}')

        items = items_response.json()['items']

      for item in items:
        versions.append(item['catalogEntry']['version'])

    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



#
# Copyright (c) 2006-2019, pkgcore contributors
# SPDX-License-Identifier: BSD-3-Clause
# Version comparison utility extracted from pkgcore and further stripped down.
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import re

from .utils import cmp
from .utils import remove_spaces

_is_gentoo_version = re.compile(
    r"^(?:\d+)(?:\.\d+)*[a-zA-Z]?(?:_(p(?:re)?|beta|alpha|rc)\d*)*$").match

suffix_regexp = re.compile("^(alpha|beta|rc|pre|p)(\\d*)$")

revision_regexp = re.compile(r".*([\.-]r\d+)")

suffix_value = {"pre": -2, "p": 1, "alpha": -4, "beta": -3, "rc": -1}
"""
gentoo ebuild version comparison
"""


def is_valid(string):
  version, _ = parse_version_and_revision(remove_spaces(string))
  return _is_gentoo_version(version)


def parse_version_and_revision(version_string):
  """
  Return a tuple of (version string, revision int) given a ``version_string``.
  """
  revision = 0
  version = version_string
  match = revision_regexp.search(version_string)
  if match:
    revision = int(match.group(1)[2:])
    version = version_string[:match.span(1)[0]]

  return version, revision


def vercmp(ver1, ver2):
  """
  Compare two versions ``ver1`` and ``ver2`` and return 0, 1, or -1 according
  to the Python 2 cmp() semantics:

      Compare the two objects x and y and return an integer according to the
      outcome. The return value is negative if x < y, zero if x == y and
      strictly positive if x > y.
  """
  if not ver1:
    if not ver2:
      return 0
    else:
      return -1
  elif not ver2:
    return 1

  ver1, rev1 = parse_version_and_revision(ver1)
  ver2, rev2 = parse_version_and_revision(ver2)

  # If the versions are the same, comparing revisions will suffice.
  if ver1 == ver2:
    # revisions are equal if 0 or None (versionless cpv)
    if not rev1 and not rev2:
      return 0
    return cmp(rev1, rev2)

  # Split up the versions into dotted strings and lists of suffixes.
  parts1 = ver1.split("_")
  parts2 = ver2.split("_")

  # If the dotted strings are equal, we can skip doing a detailed comparison.
  if parts1[0] != parts2[0]:

    # First split up the dotted strings into their components.
    ver_parts1 = parts1[0].split(".")
    ver_parts2 = parts2[0].split(".")

    # Pull out any letter suffix on the final components and keep
    # them for later.
    letters = []
    for ver_parts in (ver_parts1, ver_parts2):
      if ver_parts[-1][-1].isalpha():
        letters.append(ord(ver_parts[-1][-1]))
        ver_parts[-1] = ver_parts[-1][:-1]
      else:
        # Using -1 simplifies comparisons later
        letters.append(-1)

    # OPT: Pull length calculation out of the loop
    ver_parts1_len = len(ver_parts1)
    ver_parts2_len = len(ver_parts2)

    # Iterate through the components
    for v1, v2 in zip(ver_parts1, ver_parts2):

      # If the string components are equal, the numerical
      # components will be equal too.
      if v1 == v2:
        continue

      # If one of the components begins with a "0" then they
      # are compared as floats so that 1.1 > 1.02; else ints.
      if v1[0] != "0" and v2[0] != "0":
        v1 = int(v1)
        v2 = int(v2)
      else:
        # handle the 0.060 == 0.060 case.
        v1 = v1.rstrip("0")
        v2 = v2.rstrip("0")

      # If they are not equal, the higher value wins.
      c = cmp(v1, v2)
      if c:
        return c

    if ver_parts1_len > ver_parts2_len:
      return 1
    elif ver_parts2_len > ver_parts1_len:
      return -1

    # The dotted components were equal. Let's compare any single
    # letter suffixes.
    if letters[0] != letters[1]:
      return cmp(letters[0], letters[1])

  # The dotted components were equal, so remove them from our lists
  # leaving only suffixes.
  del parts1[0]
  del parts2[0]

  # OPT: Pull length calculation out of the loop
  parts1_len = len(parts1)
  parts2_len = len(parts2)

  # Iterate through the suffixes
  for x in range(max(parts1_len, parts2_len)):

    # If we're at the end of one of our lists, we need to use
    # the next suffix from the other list to decide who wins.
    if x == parts1_len:
      match = suffix_regexp.match(parts2[x])
      val = suffix_value[match.group(1)]
      if val:
        return cmp(0, val)
      return cmp(0, int("0" + match.group(2)))
    if x == parts2_len:
      match = suffix_regexp.match(parts1[x])
      val = suffix_value[match.group(1)]
      if val:
        return cmp(val, 0)
      return cmp(int("0" + match.group(2)), 0)

    # If the string values are equal, no need to parse them.
    # Continue on to the next.
    if parts1[x] == parts2[x]:
      continue

    # Match against our regular expression to make a split between
    # "beta" and "1" in "beta1"
    match1 = suffix_regexp.match(parts1[x])
    match2 = suffix_regexp.match(parts2[x])

    # If our int'ified suffix names are different, use that as the basis
    # for comparison.
    c = cmp(suffix_value[match1.group(1)], suffix_value[match2.group(1)])
    if c:
      return c

    # Otherwise use the digit as the basis for comparison.
    c = cmp(int("0" + match1.group(2)), int("0" + match2.group(2)))
    if c:
      return c

  # Our versions had different strings but ended up being equal.
  # The revision holds the final difference.
  return cmp(rev1, rev2)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""dpkg / Debian ecosystem helper tests."""

import requests
import vcr.unittest
import unittest
from unittest import mock
import warnings

from . import debian

from .. import cache
from .. import ecosystems


class DPKGEcosystemTest(unittest.TestCase):
  """dpkg ecosystem helper tests."""

  def test_dpkg(self):
    """Test sort_key"""
    ecosystem = debian.DPKG()
    # Debian
    self.assertGreater(
        ecosystem.sort_key('1.13.6-2'), ecosystem.sort_key('1.13.6-1'))

    # Test that <end-of-life> specifically is greater than normal versions
    self.assertGreater(
        ecosystem.sort_key('<end-of-life>'), ecosystem.sort_key('1.13.6-1'))

    # Compares base versions
    self.assertGreater(ecosystem.sort_key('1.2.3'), ecosystem.sort_key('1.2'))
    self.assertGreater(ecosystem.sort_key('1.3'), ecosystem.sort_key('1.2-3'))

    # Compares versions within the same Debian release
    self.assertGreater(
        ecosystem.sort_key('1.3+deb11u1'), ecosystem.sort_key('1.2+deb11u5'))

    # Compare versions across Debian releases
    self.assertGreater(
        ecosystem.sort_key('1.2+deb12u1'),
        ecosystem.sort_key('1.2+deb11u2'))  # deb12 > deb11
    self.assertGreater(
        ecosystem.sort_key('1.18+deb11u3'),
        ecosystem.sort_key('1.14+deb10u3'))  # 1.18 > 1.14
    self.assertGreater(
        ecosystem.sort_key('1.18+deb10'),
        ecosystem.sort_key('1.14+deb11'))  # 1.18 > 1.14

    # Echo
    self.assertGreater(
        ecosystem.sort_key("38.52.0-e1"), ecosystem.sort_key("37.52.0-e0"))
    self.assertLess(ecosystem.sort_key("453"), ecosystem.sort_key("453-e1"))
    self.assertGreater(ecosystem.sort_key("5.4.13-e1"), ecosystem.sort_key("0"))
    self.assertGreater(
        ecosystem.sort_key("1.4.0-e1"), ecosystem.sort_key("1.4.0-e0"))
    self.assertGreater(
        ecosystem.sort_key("invalid"), ecosystem.sort_key("1.4.0-e0"))
    self.assertGreater(
        ecosystem.sort_key("13.0.14.5-e1"), ecosystem.sort_key("7.64.3-e2"))
    self.assertLess(
        ecosystem.sort_key("13.0.14.5-e1"), ecosystem.sort_key("16.6-e0"))

    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('1.10.0-1'), ecosystem.sort_key('1.2.0-1'))
    self.assertLessEqual(
        ecosystem.sort_key('1.2.0-1'), ecosystem.sort_key('1.10.0-1'))

  def test_dpkg_ecosystems(self):
    """Test dpkg-based ecosystems return a DPKG ecosystem."""
    ecos = [
        'Debian',
        'Echo',
    ]
    for ecosystem_name in ecos:
      ecosystem = ecosystems.get(ecosystem_name)
      self.assertIsInstance(ecosystem, debian.DPKG)


class DebianEcosystemTest(vcr.unittest.VCRTestCase):
  """Debian ecosystem helper tests."""

  @mock.patch(
      'osv.request_helper.requests.Session.get',
      side_effect=requests.Session.get,
      autospec=True)
  @mock.patch('osv.ecosystems.debian.requests.get', side_effect=requests.get)
  def test_debian(self, first_ver_requests_mock: mock.MagicMock,
                  general_requests_mock: mock.MagicMock):
    """Test Debian ecosystem enumeration and caching behaviour"""
    in_memory_cache = cache.InMemoryCache()
    ecosystems.config.set_cache(in_memory_cache)
    ecosystem = ecosystems.get('Debian:9')

    # Tests that next version and version enumeration generally works
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('1.13.6-1', ecosystem.next_version('nginx', '1.13.5-1'))
      self.assertEqual('1.13.6-2', ecosystem.next_version('nginx', '1.13.6-1'))
      self.assertEqual('3.0.1+dfsg-2',
                       ecosystem.next_version('blender', '3.0.1+dfsg-1'))

    # Test that end-of-life enumeration is disabled
    with self.assertLogs(level='WARNING') as logs:
      self.assertEqual(
          ecosystem.enumerate_versions('nginx', '0', '<end-of-life>'), [])
    self.assertEqual(logs.output, ['WARNING:root:Package nginx has invalid fixed version: <end-of-life>. In debian release 9'])  # pylint: disable=line-too-long

    # Calls for first_version to the same ecosystem should be cached
    self.assertEqual(first_ver_requests_mock.call_count, 1)
    ecosystem.enumerate_versions('htop', '0')

    self.assertEqual(first_ver_requests_mock.call_count, 1)

    # Now start testing that Debian:10 contains different versions compared to 9
    ecosystem = ecosystems.get('Debian:10')

    # Called 2 times so far, once for nginx, once for blender.
    self.assertEqual(general_requests_mock.call_count, 3)
    # '0' as introduced version also tests the get_first_package_version func
    versions = ecosystem.enumerate_versions('cyrus-sasl2', '0', None)
    self.assertEqual(general_requests_mock.call_count, 4)

    # new ecosystem, first version requests increase by 1
    self.assertEqual(first_ver_requests_mock.call_count, 2)

    # Check that only deb10 versions are in Debian:10, and no deb9 versions
    self.assertIn('2.1.27+dfsg-1+deb10u1', versions)
    self.assertNotIn('2.1.27~101-g0780600+dfsg-3+deb9u1', versions)
    self.assertNotIn('2.1.27~101-g0780600+dfsg-3+deb9u2', versions)

    with self.assertRaises(
        ecosystems.EnumerateError), warnings.catch_warnings():
      warnings.filterwarnings('ignore', 'Avoid using this method')
      ecosystem.next_version('doesnotexist123456', '1')

    self.assertEqual(general_requests_mock.call_count, 5)

    # This should now only call the cache, and not requests.get
    ecosystem.enumerate_versions('cyrus-sasl2', '0', None)
    self.assertEqual(first_ver_requests_mock.call_count, 2)
    self.assertEqual(general_requests_mock.call_count, 5)
    ecosystems.config.set_cache(None)

  @mock.patch('osv.cache.Cache')
  def test_cache(self, cache_mock: mock.MagicMock):
    """Test caching works."""
    cache_mock.get.return_value = None
    ecosystems.config.set_cache(cache_mock)

    deb = ecosystems.get('Debian:9')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      deb.next_version('nginx', '1.13.5-1')
    cache_mock.get.assert_called_once()
    cache_mock.set.assert_called_once()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helper configurations."""
import typing

from ..cache import Cache

timeout = 30  # Timeout for HTTP(S) requests

# Used for checking out git repositories
# Intended to be set in worker.py
work_dir: typing.Optional[str] = None

shared_cache: typing.Optional[Cache] = None


def set_cache(cache: typing.Optional[Cache]):
  """Configures and enables the redis caching layer"""
  global shared_cache
  shared_cache = cache



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bioconductor ecosystem helper tests."""

import warnings
import vcr.unittest

from .. import ecosystems


class BioconductorEcosystemTest(vcr.unittest.VCRTestCase):
  """Bioconductor ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('Bioconductor')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('1.18.0', ecosystem.next_version('a4', '1.16.0'))
      self.assertEqual('1.20.0', ecosystem.next_version('a4', '1.18.0'))
      with self.assertRaises(ecosystems.EnumerateError):
        ecosystem.next_version('doesnotexist123456', '1')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: osv/importfinding.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'osv/importfinding.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17osv/importfinding.proto\x12\x03osv\x1a\x1fgoogle/protobuf/timestamp.proto\"\xbb\x01\n\rImportFinding\x12\x0e\n\x06\x62ug_id\x18\x01 \x01(\t\x12\x0e\n\x06source\x18\x02 \x01(\t\x12(\n\x08\x66indings\x18\x03 \x03(\x0e\x32\x16.osv.ImportFindingType\x12.\n\nfirst_seen\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x0clast_attempt\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp*\x8c\x04\n\x11ImportFindingType\x12\x1c\n\x18IMPORT_FINDING_TYPE_NONE\x10\x00\x12\x1f\n\x1bIMPORT_FINDING_TYPE_DELETED\x10\x01\x12$\n IMPORT_FINDING_TYPE_INVALID_JSON\x10\x02\x12\'\n#IMPORT_FINDING_TYPE_INVALID_PACKAGE\x10\x03\x12$\n IMPORT_FINDING_TYPE_INVALID_PURL\x10\x04\x12\'\n#IMPORT_FINDING_TYPE_INVALID_VERSION\x10\x05\x12&\n\"IMPORT_FINDING_TYPE_INVALID_COMMIT\x10\x06\x12%\n!IMPORT_FINDING_TYPE_INVALID_RANGE\x10\x07\x12&\n\"IMPORT_FINDING_TYPE_INVALID_RECORD\x10\x08\x12\'\n#IMPORT_FINDING_TYPE_INVALID_ALIASES\x10\t\x12(\n$IMPORT_FINDING_TYPE_INVALID_UPSTREAM\x10\n\x12\'\n#IMPORT_FINDING_TYPE_INVALID_RELATED\x10\x0b\x12\'\n#IMPORT_FINDING_TYPE_BAD_ALIASED_CVE\x10\x0c\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'osv.importfinding_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_IMPORTFINDINGTYPE']._serialized_start=256
  _globals['_IMPORTFINDINGTYPE']._serialized_end=780
  _globals['_IMPORTFINDING']._serialized_start=66
  _globals['_IMPORTFINDING']._serialized_end=253
# @@protoc_insertion_point(module_scope)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Debian ecosystem helper."""

import json
import logging
import requests

from ..third_party.univers.debian import Version as DebianVersion

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .ecosystems_base import OrderedEcosystem
from .. import cache
from ..request_helper import RequestError, RequestHelper


class DPKG(OrderedEcosystem):
  """Debian package (dpkg) ecosystem"""

  def sort_key(self, version):
    if not DebianVersion.is_valid(version):
      # If debian version is not valid, it is most likely an invalid fixed
      # version then sort it to the last/largest element
      return DebianVersion(999999, '999999')
    return DebianVersion.from_string(version)


# TODO(another-rex): Update this to use dynamically
# change depending on the project
CLOUD_API_CACHE_URL_TEMPLATE = (
    'https://storage.googleapis.com/debian-osv/first_package_output/'
    '{version}.json')
CACHE_DURATION_SECONDS = 60 * 60 * 24

debian_version_cache = cache.InMemoryCache()


class ReleaseNotFoundError(Exception):
  """Release cannot be found.

  Most likely a new release that haven't been picked up yet.

  Args:
      release_number: the release number that cannot be found.
  """
  release_number: str

  def __init__(self, release_number):
    super().__init__(release_number)
    self.release_number = release_number


@cache.cached(debian_version_cache, 24 * 60 * 60)
def _get_first_versions_for_release(release_number: str):
  """Gets the first version mapping for specific release number"""
  response = requests.get(
      CLOUD_API_CACHE_URL_TEMPLATE.format(version=release_number),
      timeout=config.timeout)
  if response.status_code == 404:
    raise ReleaseNotFoundError(release_number)

  return json.loads(response.text)


def get_first_package_version(package_name: str, release_number: str) -> str:
  """Get first package version"""

  try:
    return _get_first_versions_for_release(release_number)[package_name]
  except KeyError:
    # The package is not added when the image is first seen.
    # So it is safe to leave it as 0, indicating the earliest version
    # given by the snapshot API
    return '0'


class Debian(EnumerableEcosystem, DPKG):
  """Debian ecosystem"""

  _API_PACKAGE_URL = 'https://snapshot.debian.org/mr/package/{package}/'

  @property
  def debian_release_ver(self) -> str:
    return self.suffix if self.suffix is not None else ''

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    url = self._API_PACKAGE_URL.format(package=package.lower())
    request_helper = RequestHelper(config.shared_cache)
    try:
      text_response = request_helper.get(url)
    except RequestError as ex:
      if ex.response.status_code == 404:
        raise EnumerateError(f'Package {package} not found') from ex
      raise RuntimeError('Failed to get Debian versions for '
                         f'{package} with: {ex.response.text}') from ex

    response = json.loads(text_response)
    raw_versions: list[str] = [x['version'] for x in response['result']]

    # Remove rare cases of unknown versions
    def version_is_valid(v):
      if not DebianVersion.is_valid(v):
        logging.warning('Package %s has invalid version: %s', package, v)
        return False

      return True

    versions = [v for v in raw_versions if version_is_valid(v)]
    # Sort to ensure it is in the correct order
    self.sort_versions(versions)
    # The only versions with +deb
    versions = [
        x for x in versions
        if '+deb' not in x or f'+deb{self.debian_release_ver}' in x
    ]

    if introduced == '0':
      # Update introduced to the first version of the debian version
      introduced = get_first_package_version(package, self.debian_release_ver)

    if fixed is not None and not DebianVersion.is_valid(fixed):
      logging.warning(
          'Package %s has invalid fixed version: %s. In debian release %s',
          package, fixed, self.debian_release_ver)
      return []

    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""pubsub tests."""
import unittest
from unittest import mock

from . import pubsub


class PubsubTest(unittest.TestCase):
  """Pub/Sub tests."""

  @mock.patch('google.cloud.pubsub_v1.PublisherClient')
  @mock.patch('osv.utils.get_google_cloud_project')
  def test_publish_failure(self, mock_get_project, mock_publisher_client):
    """Test publish_failure."""
    mock_get_project.return_value = 'test-project'
    mock_publisher = mock.MagicMock()
    mock_publisher_client.return_value = mock_publisher

    pubsub.publish_failure(b'test data', attr1='value1')

    mock_publisher.topic_path.assert_called_once_with('test-project',
                                                      'failed-tasks')
    topic_path = mock_publisher.topic_path.return_value
    mock_publisher.publish.assert_called_once_with(
        topic_path, b'test data', attr1='value1')

  @mock.patch('osv.utils.get_google_cloud_project')
  def test_publish_failure_no_project(self, mock_get_project):
    """Test publish_failure with no project."""
    mock_get_project.return_value = ''
    with self.assertRaises(RuntimeError), self.assertLogs() as cm:
      pubsub.publish_failure(b'test data')

    self.assertEqual(
        ['ERROR:root:GOOGLE_CLOUD_PROJECT not set, cannot send retry message'],
        cm.output)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bug helpers."""

import enum
import re


class BugStatus(enum.IntFlag):
  """Bug status."""

  UNPROCESSED = 0
  PROCESSED = 1
  INVALID = 2


# Groups of numbers, or {RC,alpha,beta,preview}<optional number>. Include a
# negative lookbehind to avoid catching cases like "arc" which may be part of a
# project name.
VERSION_COMPONENT_REGEX = re.compile(
    r'(\d+|(?<![a-z])(?:rc|alpha|beta|preview)\d*)', re.IGNORECASE)


def normalize_tag(tag):
  """Normalize a single tag for fuzzy version searching."""
  components = VERSION_COMPONENT_REGEX.findall(tag)
  return '-'.join(components)


def normalize_tags(tags):
  """Normalize tags for fuzzy version searching."""
  return [normalize_tag(tag) for tag in tags]


def populate_indices(bug):
  """Write search indices for the bug."""
  bug['has_affected'] = bool(bug.get('affected'))

  search_indices = []
  project = bug.get('project')
  if project:
    search_indices.append(project)

  search_indices.append(bug.key.id_or_name)
  search_indices.extend(bug.key.id_or_name.split('-'))
  bug['search_indices'] = search_indices



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test helpers."""
import contextlib
import datetime
import difflib
import os
import pprint
import signal
import time

from proto import datetime_helpers
import requests
import subprocess
import tempfile
from unittest import mock

import pygit2
import pygit2.enums

from . import gcs_mock

_EMULATOR_TIMEOUT = 30
_DATASTORE_EMULATOR_PORT = '8002'
TEST_PROJECT_ID = 'test-osv'


def ExpectationTest(test_data_dir):  # pylint: disable=invalid-name
  """Mixin for test output generation/comparison."""

  class Mixin:
    """Mixin."""

    def _load_expected(self, expected_name, actual):
      """Load expected data."""
      expected_path = os.path.join(
          test_data_dir, f'{self.__class__.__name__}_{expected_name}.txt')
      if os.getenv('TESTS_GENERATE'):
        pp = pprint.PrettyPrinter(indent=4)
        with open(expected_path, 'w') as f:
          f.write(pp.pformat(actual))

      with open(expected_path) as f:
        eval_globals = globals()
        eval_globals['call'] = mock.call
        eval_globals['DatetimeWithNanoseconds'] = datetime_helpers.DatetimeWithNanoseconds  # pylint: disable=line-too-long
        return eval(f.read(), eval_globals)  # pylint: disable=eval-used

    def expect_dict_equal(self, expected_name, actual):
      """Check if the output dict is equal to the expected value."""
      self.assertDictEqual(self._load_expected(expected_name, actual), actual)

    def expect_lines_equal(self, expected_name, actual_lines):
      """Check if the output lines is equal to the expected value,
      printing a diff when it is different."""
      expected_lines = self._load_expected(expected_name, actual_lines)
      if expected_lines != actual_lines:
        diff = difflib.unified_diff(expected_lines, actual_lines, 'expected',
                                    'actual')
        print(''.join(diff))
        self.fail()

    def expect_equal(self, expected_name, actual):
      """Check if the output is equal to the expected value."""
      self.assertEqual(self._load_expected(expected_name, actual), actual)

  return Mixin


class MockRepo:
  """Mock repo."""

  def __init__(self, path):
    self.path = path
    self._repo = pygit2.init_repository(path, True)
    tree = self._repo.TreeBuilder().write()
    author = pygit2.Signature('OSV', 'infra@osv.dev')
    self._repo.create_commit('HEAD', author, author, 'Initial commit', tree, [])

  def add_file(self, path, contents):
    """Adds a file."""
    oid = self._repo.write(pygit2.enums.ObjectType.BLOB, contents)
    self._repo.index.add(
        pygit2.IndexEntry(path, oid, pygit2.enums.FileMode.BLOB))
    self._repo.index.write()

  def delete_file(self, path):
    """Delete a file."""
    self._repo.index.remove(path)
    self._repo.index.write()

  def commit(self, author_name, author_email, message='Changes'):
    """Makes a commit."""
    tree = self._repo.index.write_tree()
    author = pygit2.Signature(author_name, author_email)
    self._repo.create_commit('HEAD', author, author, message, tree,
                             [self._repo.head.peel().id])


class _DatastoreEmulator:

  def __init__(self, url):
    self._url = url

  def reset(self):
    """Clear emulator data."""
    resp = requests.post(f'http://{self._url}/reset', timeout=_EMULATOR_TIMEOUT)
    resp.raise_for_status()


def _is_emulator_installed():
  """Check if the gcloud firestore emulator is installed."""
  try:
    # We check if the component is installed.
    output = subprocess.check_output([
        'gcloud', 'components', 'list', '--quiet',
        '--filter=id:cloud-firestore-emulator', '--format=value(state.name)'
    ],
                                     text=True,
                                     stderr=subprocess.DEVNULL)
    return 'Not Installed' not in output
  except (subprocess.CalledProcessError, FileNotFoundError):
    return False


@contextlib.contextmanager
def datastore_emulator():
  """A context for running the datastore emulator in."""
  if not _is_emulator_installed():
    raise RuntimeError(
        'gcloud firestore emulator is not installed or not working. '
        'Please install it by running '
        '`gcloud components install cloud-firestore-emulator` and/or '
        'ensure `gcloud` is in your PATH.')

  port = os.environ.get('DATASTORE_EMULATOR_PORT', _DATASTORE_EMULATOR_PORT)
  env = {
      'DATASTORE_EMULATOR_HOST': 'localhost:' + port,
      'DATASTORE_PROJECT_ID': TEST_PROJECT_ID,
      'GOOGLE_CLOUD_PROJECT': TEST_PROJECT_ID,
  }
  old_env = {}
  for k, v in env.items():
    old_env[k] = os.environ.get(k)
    os.environ[k] = v

  url = f'localhost:{port}'
  # Try shutdown an existing emulator on the port
  try:
    requests.post(f'http://{url}/shutdown', timeout=_EMULATOR_TIMEOUT)
  except requests.ConnectionError:
    pass

  # Start the emulator in a new process group so we can terminate the
  # subprocesses it spawns.
  emu = subprocess.Popen([
      'gcloud',
      'emulators',
      'firestore',
      'start',
      '--database-mode=datastore-mode',
      f'--host-port={url}',
  ],
                         stdout=subprocess.DEVNULL,
                         stderr=subprocess.DEVNULL,
                         start_new_session=True)

  # Wait for ready
  start = time.time()
  while time.time() - start < _EMULATOR_TIMEOUT:
    try:
      resp = requests.get(f'http://{url}', timeout=_EMULATOR_TIMEOUT)
      if resp.ok:
        break
    except requests.ConnectionError:
      pass
    time.sleep(0.5)
  else:
    os.killpg(emu.pid, signal.SIGKILL)
    raise RuntimeError('Datastore emulator did not get ready in time.')

  # Also mock the GCS bucket
  with gcs_mock.gcs_mock():
    yield _DatastoreEmulator(url)

  # Stop the process
  requests.post(f'http://{url}/shutdown', timeout=_EMULATOR_TIMEOUT)
  try:
    os.killpg(emu.pid, signal.SIGTERM)
    emu.wait(timeout=_EMULATOR_TIMEOUT)
  except subprocess.TimeoutExpired:
    print('Emulator ignored SIGTERM. Sending SIGKILL')
    os.killpg(emu.pid, signal.SIGKILL)
    emu.wait()
  finally:
    # restore environment variables
    for k, v in old_env.items():
      if v is None:
        os.environ.pop(k, None)
      else:
        os.environ[k] = v


def mock_datetime(test):
  """Mocks datetime."""
  for to_mock in ('osv.models.utcnow', 'osv.utcnow'):
    patcher = mock.patch(to_mock)
    mock_utcnow = patcher.start()
    mock_utcnow.return_value = datetime.datetime(
        2021, 1, 1, tzinfo=datetime.UTC)
    test.addCleanup(patcher.stop)


def mock_repository(test):
  """Creates a mock repo."""
  tmp_dir = tempfile.TemporaryDirectory()
  test.remote_source_repo_path = tmp_dir.name
  test.addCleanup(tmp_dir.cleanup)
  return MockRepo(test.remote_source_repo_path)


def mock_clone(test, func=None, return_value=None):
  """Mocks clone_repository."""
  patcher = mock.patch('osv.repos.clone')
  mocked = patcher.start()
  if return_value:
    mocked.return_value = return_value
  else:
    mocked.side_effect = func
  test.addCleanup(patcher.stop)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helpers tests."""

import json

import re

import unittest

from . import _ecosystems


class EcosystemsTest(unittest.TestCase):
  """Ecosystem helpers tests."""

  def setUp(self):
    with open('osv/osv-schema/validation/schema.json') as schema_f:
      schema = json.load(schema_f)
    self.schema_ecosystems = re.compile(
        schema['$defs']['ecosystemWithSuffix']['pattern'])
    self.canonical_ecosystems = _ecosystems._ecosystems.keys()  # pylint: disable=protected-access

  def test_ecosystem_supported_by_schema(self):
    """Test ecosystems referenced exist in schema definition"""
    for ecosystem in self.canonical_ecosystems:
      self.assertIsNotNone(
          self.schema_ecosystems.match(ecosystem),
          msg=f'"{ecosystem}" not defined in schema')

    for ecosystem in _ecosystems.package_urls:
      self.assertIsNotNone(
          self.schema_ecosystems.match(ecosystem),
          msg=f'Purl "{ecosystem}" not defined in schema')

    for ecosystem in _ecosystems._OSV_TO_DEPS_ECOSYSTEMS_MAP:  # pylint: disable=protected-access
      self.assertIsNotNone(
          self.schema_ecosystems.match(ecosystem),
          msg=f'"{ecosystem}" in deps.dev map not defined in schema')


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pub/Sub helpers."""
import logging

from google.cloud import pubsub_v1

from . import utils

FAILED_TASKS_TOPIC = 'failed-tasks'
_pubsub_client = None


def _get_pubsub_client() -> pubsub_v1.PublisherClient:
  """Get a Pub/Sub publisher client."""
  global _pubsub_client
  if _pubsub_client is None:
    _pubsub_client = pubsub_v1.PublisherClient()
  return _pubsub_client


def publish_failure(data: bytes, **attributes: str):
  """Publishes a message to the failed-tasks topic."""
  project = utils.get_google_cloud_project()
  if not project:
    logging.error('GOOGLE_CLOUD_PROJECT not set, cannot send retry message')
    raise RuntimeError('GOOGLE_CLOUD_PROJECT not set')

  publisher = _get_pubsub_client()
  topic = publisher.topic_path(project, FAILED_TASKS_TOPIC)

  try:
    publisher.publish(topic, data, **attributes)
    logging.info('Published failure message to %s with attributes %s', topic,
                 attributes)
  except Exception:
    logging.exception('Failed to publish failure message to %s', topic)
    raise



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and
# https://github.com/nexB/univers for support and download.

# Modified to support revisions in the form of <ver>-r<rev-num>
import attr
import re

from . import gentoo
from .utils import remove_spaces


class InvalidVersion(ValueError):
  pass


def is_valid_alpine_version(s: str):
  """
  Return True is the string `s` is a valid Alpine version.
  We do not support yet version strings that start with
  non-significant zeros.
  For example:
  >>> is_valid_alpine_version("006")
  False
  >>> is_valid_alpine_version("1.2.3")
  True
  >>> is_valid_alpine_version("02-r1")
  True
  """
  search = AlpineLinuxVersion.version_extractor.search(s)
  if not search:
    return False

  s = search.group(1)
  left, _, _ = s.partition(".")
  # handle the suffix case
  left, _, _ = left.partition("-")
  if not left.isdigit():
    return True
  i = int(left)
  return str(i) == left


@attr.s(frozen=True, order=False, hash=True)
class Version:
  """
  Base version mixin to subclass for each version syntax implementation.
  Each version subclass is:
  - immutable and hashable
  - comparable and orderable e.g., such as implementing all rich comparison
    operators or implementing functools.total_ordering. The default is to
    compare the value as-is.
  """

  # the original string used to build this Version
  string = attr.ib(type=str)

  # the normalized string for this Version, stored without spaces and
  # lowercased. Any leading v is removed too.
  normalized_string = attr.ib(type=str, default=None, repr=False)

  # a comparable scheme-specific version object constructed from
  # the version string
  value = attr.ib(default=None, repr=False, type=(str, int))

  def __attrs_post_init__(self):
    normalized_string = self.normalize(self.string)
    if not self.is_valid(normalized_string):
      raise InvalidVersion(f"{self.string!r} is not a valid {self.__class__!r}")

    # Set the normalized string as default value

    # Notes: setattr is used because this is an immutable frozen instance.
    # See https://www.attrs.org/en/stable/init.html?#post-init
    object.__setattr__(self, "normalized_string", normalized_string)
    value = self.build_value(normalized_string)
    object.__setattr__(self, "value", value)

  @classmethod
  def is_valid(cls, string):
    """
    Return True if the ``string`` is a valid version for its scheme or False
    if not valid. The empty string, None, False and 0 are considered invalid.
    Subclasses should implement this.
    """
    return bool(string)

  @classmethod
  def normalize(cls, string):
    """
    Return a normalized version string from ``string ``. Subclass can override.
    """
    # FIXME: Is removing spaces and strip v the right thing to do?
    return remove_spaces(string).rstrip("v ").strip()

  @classmethod
  def build_value(cls, string):
    """
    Return a wrapped version "value" object for a version ``string``.
    Subclasses can override. The default is a no-op and returns the string
    as-is, and is called by default at init time with the computed
    normalized_string.
    """
    return string

  def satisfies(self, constraint):
    """
    Return True is this Version satisfies the ``constraint``
    VersionConstraint. Satisfying means that this version is "within" the
    ``constraint``.
    """
    return self in constraint

  def __str__(self):
    return str(self.value)

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.value.__eq__(other.value)

  def __lt__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.value.__lt__(other.value)

  def __gt__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.value.__gt__(other.value)

  def __le__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.value.__le__(other.value)

  def __ge__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.value.__ge__(other.value)


@attr.s(frozen=True, order=False, eq=False, hash=True)
class AlpineLinuxVersion(Version):
  """Alpine linux version"""
  # E.g. For this version (1.9.5_p2-r3), the following regex
  # extracts (1.9.5_p2) and (3)
  version_extractor = re.compile(r'(.+?)(?:[\.-]r(\d+))?$')

  # Some suffixes are not separated with an underscore. E.g. 1.9.5p2
  # This should find them for inserting an underscore (replace with r'\1_\2')
  # See: https://gitlab.alpinelinux.org/alpine/abuild/-/issues/10088 for more context
  patch_finder = re.compile(r'(\d+)(p\d+)')

  @classmethod
  def add_underscore(cls, input: str) -> str:
    return re.sub(cls.patch_finder, r'\1_\2', input, count=1)

  @classmethod
  def build_value(cls, string: str):
    search = cls.version_extractor.search(string)
    main_group = search.group(1)
    main_group_patched = cls.add_underscore(main_group)
    return (main_group_patched, int(search.group(2) or 0))

  @classmethod
  def is_valid(cls, string):
    string_patched = cls.add_underscore(string)
    return is_valid_alpine_version(string_patched) and gentoo.is_valid(
        string_patched)

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    gentoo_vercmp = gentoo.vercmp(self.value[0], other.value[0])
    return gentoo_vercmp == 0 and self.value[1] == other.value[1]

  def __lt__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented

    gentoo_vercmp = gentoo.vercmp(self.value[0], other.value[0])
    if gentoo_vercmp == 0:
      return self.value[1] < other.value[1]

    return gentoo_vercmp < 0

  def __gt__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented

    gentoo_vercmp = gentoo.vercmp(self.value[0], other.value[0])
    if gentoo_vercmp == 0:
      return self.value[1] > other.value[1]
    return gentoo_vercmp > 0

  def __le__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return not self.__gt__(other)

  def __ge__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return not self.__lt__(other)



#!/usr/bin/env python3

import json
import shutil

MARKDOWN_TABLE_MARKER_START = '<!-- begin auto-generated ecosystems list -->'
MARKDOWN_TABLE_MARKER_END = '<!-- end auto-generated ecosystems list -->'

# ensure that the ecosystems are sorted alphabetically and don't have extra whitespace
ecosystems: dict[str, str] = {
  k: v.strip() for k, v in sorted(
    json.loads(open('ecosystems.json').read()).items(),
    key=lambda item: item[0].casefold()
  )
}

# write back to the json file in case there were any changes
open('ecosystems.json', 'w').write(json.dumps(ecosystems, indent=2) + '\n')


def update_json_schema(filename: str):
  """
  Updates references to ecosystems defined in the OSV JSON schema
  :return:
  """
  schema = json.loads(open(filename).read())

  names = ecosystems.keys()
  pattern = ''

  for name in names:
    pattern += name.replace('.', '\\.')
    pattern += '|'

  # this is a special "ecosystem" name
  pattern += 'GIT'

  schema['$defs']['ecosystemName']['enum'] = list(names)
  schema['$defs']['ecosystemWithSuffix']['pattern'] = f'^({pattern})(:.+)?$'

  open(filename, 'w').write(json.dumps(schema, indent=2) + '\n')


def generate_ecosystems_markdown_table() -> str:
  """
  Generates a Markdown table of supported ecosystems with descriptions
  :return:
  """
  table = '| Ecosystem | Description |\n'
  table += '|-----------|-------------|\n'

  for name, description in ecosystems.items():
    table += f'| `{name}` | {description} |\n'

  return table


def update_schema_md():
  """
  Updates the schema.md file with the list of ecosystems
  :return:
  """
  md = open('docs/schema.md').read()

  table_start_index = md.index(MARKDOWN_TABLE_MARKER_START)
  table_end_index = md.index(MARKDOWN_TABLE_MARKER_END)

  assert table_start_index < table_end_index, 'Table start index must be before table end index'

  table = generate_ecosystems_markdown_table()
  table += '| Your ecosystem here. | [Send us a PR](https://github.com/ossf/osv-schema/compare). |\n'

  md = '{0}{1}\n\n{2}\n{3}{4}'.format(
    md[:table_start_index],
    MARKDOWN_TABLE_MARKER_START,
    table,
    MARKDOWN_TABLE_MARKER_END,
    md[table_end_index + len(MARKDOWN_TABLE_MARKER_END):]
  )

  open('docs/schema.md', 'w').write(md)


def convert_to_go_constant_name(name: str) -> str:
  """
  Converts the "human" name of an ecosystem to a Go constant name, mostly
  by removing spaces and dashes and converting to PascalCase.

  Some ecosystems have special cases, like "crates.io" which is converted to "CratesIO".

  :param name:
  :return:
  """
  if name == 'crates.io':
    return 'EcosystemCratesIO'

  if name == 'npm':
    return 'EcosystemNPM'

  name = name[0].upper() + name[1:]
  name = name.replace('-', '').replace(' ', '')

  return f'Ecosystem{name}'


def generate_ecosystems_go_constants() -> str:
  """
  Generates a list of Go constants, with a constant per ecosystem
  :return:
  """

  constants = list(map(lambda x: (convert_to_go_constant_name(x), x), ecosystems.keys()))
  longest = max(map(lambda x: len(x[0]), constants))

  code = 'const (\n'
  for constant, name in constants:
    code += f'\t{constant.ljust(longest)} Ecosystem = "{name}"\n'
  code += ')\n'

  return code


def update_go_constants():
  """
  Updates the Go constants with the list of ecosystems
  :return:
  """
  go = open('bindings/go/osvschema/constants.go').read()

  ecosystem_constants_start = go.index('type Ecosystem string\n\nconst (\n')
  ecosystems_constants_end = go.index(')', ecosystem_constants_start)

  go = '{0}{1}\n\n{2}{3}'.format(
    go[:ecosystem_constants_start],
    'type Ecosystem string',
    generate_ecosystems_go_constants(),
    go[ecosystems_constants_end + 2:]
  )

  open('bindings/go/osvschema/constants.go', 'w').write(go)


update_go_constants()
update_json_schema('validation/schema.json')
shutil.copy('validation/schema.json', 'tools/osv-linter/internal/checks/schema_generated.json')
update_schema_md()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""APK / Alpine ecosystem helper tests."""

import os
import unittest
from unittest import mock
import warnings

from . import alpine

from .. import cache
from .. import ecosystems
from .. import repos


class APKEcosystemTest(unittest.TestCase):
  """APK ecosystem helper tests."""

  def test_apk(self):
    """Test APK ecosystem sort_key behaviour"""
    ecosystem = alpine.APK()

    # Alpine
    # Should not throw exception
    ecosystem.sort_key('1.9.5p2')
    ecosystem.sort_key('1.9.5p2-r0')

    self.assertGreater(
        ecosystem.sort_key('1.9.5p3'), ecosystem.sort_key('1.9.5p2'))
    self.assertGreater(
        ecosystem.sort_key('1.9.5p1'), ecosystem.sort_key('1.9.5'))

    self.assertGreater(
        ecosystem.sort_key('1.13.2-r0'), ecosystem.sort_key('1.13.2_alpha'))

    # Check invalid version handle
    self.assertGreater(
        ecosystem.sort_key('1-0-0'), ecosystem.sort_key('1.13.2-r0'))
    self.assertEqual(
        ecosystem.sort_key('1.13.2-r0'), ecosystem.sort_key('1.13.2-r0'))

    # Chainguard
    self.assertGreater(
        ecosystem.sort_key('38.52.0-r0'), ecosystem.sort_key('37.52.0-r0'))
    self.assertLess(ecosystem.sort_key('453'), ecosystem.sort_key('453-r1'))
    self.assertGreater(ecosystem.sort_key('5.4.13-r1'), ecosystem.sort_key('0'))
    self.assertGreater(
        ecosystem.sort_key('1.4.0-r1'), ecosystem.sort_key('1.4.0-r0'))
    self.assertGreater(
        ecosystem.sort_key('invalid'), ecosystem.sort_key('1.4.0-r0'))

    # MinimOS
    self.assertGreater(
        ecosystem.sort_key('38.52.0-r0'), ecosystem.sort_key('37.52.0-r0'))
    self.assertLess(ecosystem.sort_key('453'), ecosystem.sort_key('453-r1'))
    self.assertGreater(ecosystem.sort_key('5.4.13-r1'), ecosystem.sort_key('0'))
    self.assertGreater(
        ecosystem.sort_key('1.4.0-r1'), ecosystem.sort_key('1.4.0-r0'))
    self.assertGreater(
        ecosystem.sort_key('invalid'), ecosystem.sort_key('1.4.0-r0'))
    self.assertGreater(
        ecosystem.sort_key('13.0.14.5-r1'), ecosystem.sort_key('7.64.3-r2'))
    self.assertLess(
        ecosystem.sort_key('13.0.14.5-r1'), ecosystem.sort_key('16.6-r0'))

    # Check >= / <= methods: https://github.com/google/osv.dev/pull/3978
    self.assertGreaterEqual(
        ecosystem.sort_key('1.10.0-r0'), ecosystem.sort_key('1.2.0-r0'))
    self.assertLessEqual(
        ecosystem.sort_key('1.2.0-r0'), ecosystem.sort_key('1.10.0-r0'))

  def test_apk_ecosystems(self):
    """Test apk-based ecosystems return an APK ecosystem."""
    ecos = [
        'Alpine',
        'Alpaquita',
        'BellSoft Hardened Containers',
        'Chainguard',
        'MinimOS',
        'Wolfi',
    ]
    for ecosystem_name in ecos:
      ecosystem = ecosystems.get(ecosystem_name)
      self.assertIsInstance(ecosystem, alpine.APK)


class AlpineEcosystemTest(unittest.TestCase):
  """Alpine ecosystem helper tests."""
  _TEST_DATA_DIR = os.path.join(
      os.path.dirname(os.path.abspath(__file__)), 'testdata')

  @mock.patch(
      'osv.repos.ensure_updated_checkout',
      side_effect=repos.ensure_updated_checkout)
  def test_alpine(self, ensure_updated_checkout_mock: mock.MagicMock):
    """Test Alpine ecosystem enumeration and caching behaviour"""
    in_memory_cache = cache.InMemoryCache()
    ecosystems.config.set_cache(in_memory_cache)

    # Set work_dir to allow cloning/fetching
    ecosystems.config.work_dir = self._TEST_DATA_DIR
    ecosystem = ecosystems.get('Alpine:v3.16')
    self.assertEqual(ensure_updated_checkout_mock.call_count, 0)
    # Tests that next version and version enumeration generally works
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('1.12.2-r1', ecosystem.next_version('nginx', '1.12.2'))
      self.assertEqual(ensure_updated_checkout_mock.call_count, 1)
      self.assertEqual('1.16.1-r0',
                       ecosystem.next_version('nginx', '1.16.0-r4'))
      # Second call should use cache, so call count should not increase
      self.assertEqual(ensure_updated_checkout_mock.call_count, 1)

      # Check letter suffixes clone correctly
      self.assertEqual('2.78c-r0',
                       ecosystem.next_version('blender', '2.78a-r1'))

    ecosystems.config.set_cache(None)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Packagist ecosystem helper."""

import json
import re
from typing import List

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from ..request_helper import RequestError, RequestHelper


class PackagistVersion:
  # pylint: disable=line-too-long
  """
  Follows the packagist version ordering, which is recommended to be semver, but
  not enforced to be semver. The php standard version comparison code written in
  C is located here:
  https://github.com/php/php-src/blob/master/ext/standard/versioning.c

  The function first replaces _, - and + with a dot . in the version strings and
  also inserts dots . before and after any non number so that for example
  '4.3.2RC1' becomes '4.3.2.RC.1'. Then it compares the parts starting from
  left to right.

  If a part contains special version strings these are handled in the following
  order:
  any string not found in this list < dev < alpha = a < beta = b < RC = rc < # < pl = p.
  This way not only versions with different levels like '4.1' and '4.1.2' can be
  compared but also any PHP specific version containing development state.

  ---

  ## Known differences:
  The following are some examples of known differences between this python
  implementation and the C implementation of PHP

  - In this version, special version strings need to exactly match to not be
    considered "any other string", while in the original implementation the
    string only need to start with one of the listed strings.
  """
  # pylint: enable=line-too-long

  version_str: str
  canonicalized_version: str

  def __init__(self, version: str):
    self.version_str = version
    self.canonicalized_version = self.php_canonicalize_version(version)

  def __str__(self) -> str:
    return self.version_str

  def __hash__(self):
    return self.canonicalized_version

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented
    return self.canonicalized_version == other.canonicalized_version

  def __lt__(self, other):
    return self.__cmp__(other) < 0

  def __le__(self, other):
    return self.__cmp__(other) <= 0

  def __gt__(self, other):
    return self.__cmp__(other) > 0

  def __ge__(self, other):
    return self.__cmp__(other) >= 0

  def __cmp__(self, other):
    return self.php_version_compare(self.version_str, other.version_str)

  @staticmethod
  def php_slices_compare(a_split: List[str], b_split: List[str]):
    """
    Compare php versions after being split by '.'
    """
    for a, b in zip(a_split, b_split):
      if a.isdigit() and b.isdigit():
        compare = int(a) - int(b)
      elif not a.isdigit() and not b.isdigit():
        compare = PackagistVersion.compare_special_versions(a, b)
      elif a.isdigit():
        compare = PackagistVersion.compare_special_versions('#', b)
      else:
        compare = PackagistVersion.compare_special_versions(a, '#')

      if compare != 0:
        if compare > 0:
          return 1
        return -1

    if len(a_split) > len(b_split):
      next_char = a_split[len(b_split)]
      if next_char.isdigit():
        return 1
      return PackagistVersion.php_slices_compare(a_split[len(b_split):], ['#'])

    if len(a_split) < len(b_split):
      next_char = b_split[len(a_split)]
      if next_char.isdigit():
        return -1
      return PackagistVersion.php_slices_compare(['#'], b_split[len(a_split):])

    return 0

  @staticmethod
  def php_version_compare(version_a: str, version_b: str) -> int:
    """
    Given two php versions, compare which is newer

    :return: 1 if a > b, -1 if b > a, 0 if a == b
    """
    version_a = PackagistVersion.php_canonicalize_version(version_a)
    version_b = PackagistVersion.php_canonicalize_version(version_b)

    a_split = version_a.split('.')
    b_split = version_b.split('.')
    return PackagistVersion.php_slices_compare(a_split, b_split)

  @staticmethod
  def php_canonicalize_version(version: str) -> str:
    """
    Replaces special separators (`-`,`_`,`+`) with `.`, and inserts `.`
    between any digit and non-digit.
    """
    if version.startswith('v'):
      version = version[1:]
    replaced = re.sub('[-_+]', '.', version)
    replaced = re.sub(r'([^\d.])(\d)', r'\1.\2', replaced)
    replaced = re.sub(r'(\d)([^\d.])', r'\1.\2', replaced)
    return replaced

  SPECIAL_CHARACTER_ORDER = {
      "dev": 0,
      "alpha": 1,
      "a": 1,
      "beta": 2,
      "b": 2,
      "RC": 3,
      "rc": 3,
      "#": 4,
      "pl": 5,
      "p": 5,
      None: 0,
  }

  @staticmethod
  def compare_special_versions(version_part_a: str, version_part_b: str) -> int:
    # pylint: disable=line-too-long
    """
    Compares the order of special characters against the order specified in php
    docs.

    any string not found in this list < dev < alpha = a < beta = b < RC = rc < # < pl = p.

    :return: 1 if a > b, -1 if b > a, 0 if a == b
    """
    # pylint: enable=line-too-long
    # This isn't quite the behaviour of the c implementation of php
    # version_compare
    # In php if the part starts with special_chars its enough.
    # ### For example:
    # *PHP implementation:*
    # `1.0.0beta1 = 1.0.0betawithsomeothertext1`
    #
    # *This python implementation*
    # `1.0.0beta1 > 1.0.0betawithsomeothertext1`
    # Because "any string not found in this list" should apply to
    # `betawithsomeothertext`
    found_a = PackagistVersion.SPECIAL_CHARACTER_ORDER.get(version_part_a, -1)
    found_b = PackagistVersion.SPECIAL_CHARACTER_ORDER.get(version_part_b, -1)

    if found_a > found_b:
      return 1
    if found_a < found_b:
      return -1
    return 0


class Packagist(EnumerableEcosystem):
  """Packagist ecosystem"""

  _API_PACKAGE_URL = 'https://repo.packagist.org/p2/{package}.json'

  def sort_key(self, version):
    return PackagistVersion(version)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    url = self._API_PACKAGE_URL.format(package=package.lower())
    request_helper = RequestHelper(config.shared_cache)
    try:
      text_response = request_helper.get(url)
    except RequestError as ex:
      if ex.response.status_code == 404:
        raise EnumerateError(f'Package {package} not found') from ex
      raise RuntimeError('Failed to get Packagist versions for '
                         f'{package} with: {ex.response.text}') from ex

    response = json.loads(text_response)
    versions: list[str] = [x['version'] for x in response['packages'][package]]
    self.sort_versions(versions)
    # TODO(rexpan): Potentially filter out branch versions like dev-master

    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Impact tests."""

import codecs
import unittest
from unittest import mock

from . import impact
from . import tests
from . import models

from google.cloud import ndb


class UpdateAffectedCommitsTests(unittest.TestCase):
  """update_affected_commits tests."""

  def setUp(self):
    self._ds_emulator.reset()

  @classmethod
  def setUpClass(cls):
    # Start the emulator BEFORE creating the ndb client
    cls._ds_emulator = cls.enterClassContext(tests.datastore_emulator())
    cls.enterClassContext(ndb.Client().context(cache_policy=False))

  def test_update_single_page(self):
    """Test update_affected_commits with a single page."""
    commits = {
        '00',
        '01',
        '02',
    }

    impact.update_affected_commits('BUG-1', commits, True)
    affected_commits = list(models.AffectedCommits.query())
    self.assertEqual(1, len(affected_commits))
    affected_commits = affected_commits[0]

    self.assertEqual('BUG-1-0', affected_commits.key.id())
    self.assertEqual('BUG-1', affected_commits.bug_id)
    self.assertEqual(0, affected_commits.page)
    self.assertTrue(affected_commits.public)

    self.assertCountEqual([
        b'00',
        b'01',
        b'02',
    ], [codecs.encode(commit, 'hex') for commit in affected_commits.commits])

  def test_update_single_page_nonpublic(self):
    """Test update_affected_commits with a single page."""
    commits = {
        '00',
        '01',
        '02',
    }

    impact.update_affected_commits('BUG-1', commits, False)
    affected_commits = list(models.AffectedCommits.query())
    self.assertEqual(1, len(affected_commits))
    affected_commits = affected_commits[0]

    self.assertEqual('BUG-1-0', affected_commits.key.id())
    self.assertEqual('BUG-1', affected_commits.bug_id)
    self.assertEqual(0, affected_commits.page)
    self.assertFalse(affected_commits.public)

    self.assertCountEqual([
        b'00',
        b'01',
        b'02',
    ], [codecs.encode(commit, 'hex') for commit in affected_commits.commits])

  def test_update_multiple_pages(self):
    """Test update_affected_commits with multiple page."""
    # The emulator cannot handle indexing 10000 commits in a single entity.
    with mock.patch.object(models.AffectedCommits, 'MAX_COMMITS_PER_ENTITY',
                           100):
      # These pre-populated pages should be deleted.
      for i in range(10):
        models.AffectedCommits(
            id=f'BUG-1-{i}', bug_id='BUG-1', commits=[], public=True,
            page=i).put()

      # These should not be deleted as they're for a different bug.
      models.AffectedCommits(
          id='BUG-2-0', bug_id='BUG-2', commits=[], public=True, page=0).put()

      commits = {'%08d' % number for number in range(260)}

      impact.update_affected_commits('BUG-1', commits, True)
      affected_commits = list(models.AffectedCommits.query())
      self.assertEqual(4, len(affected_commits))

      # Check that the unrelated entry still exists.
      self.assertTrue(any(c.key.id() == 'BUG-2-0') for c in affected_commits)

      # Check that the new pages got written properly.
      affected_commits = list(
          models.AffectedCommits.query(
              models.AffectedCommits.bug_id == 'BUG-1'))

      for i, result in enumerate(affected_commits):
        self.assertEqual(f'BUG-1-{i}', result.key.id())
        self.assertEqual('BUG-1', result.bug_id)
        self.assertEqual(i, result.page)
        self.assertTrue(result.public)

      self.assertCountEqual(
          [b'%08d' % i for i in range(100)],
          [codecs.encode(c, 'hex') for c in affected_commits[0].commits])
      self.assertCountEqual(
          [b'%08d' % i for i in range(100, 200)],
          [codecs.encode(c, 'hex') for c in affected_commits[1].commits])
      self.assertCountEqual(
          [b'%08d' % i for i in range(200, 260)],
          [codecs.encode(c, 'hex') for c in affected_commits[2].commits])

  def test_update_no_commits(self):
    """Test updates with no commits."""
    # These pre-populated pages should be deleted.
    for i in range(10):
      models.AffectedCommits(
          id=f'BUG-1-{i}', bug_id='BUG-1', commits=[], public=True,
          page=i).put()

    impact.update_affected_commits('BUG-1', set(), True)
    affected_commits = list(models.AffectedCommits.query())
    self.assertEqual(0, len(affected_commits))



"""Test parsing CSAF v2 advisories"""
import unittest

from redhat_osv.csaf import Remediation


class CSAFTest(unittest.TestCase):
    """class to handle remediation advice in CSAF data"""

    def test_parse_remediation(self):
        """Test parsing a CSAF Remediation and unpacking cpe and purl data"""
        cpe = "cpe:/a:redhat:rhel_tus:8.4::appstream"
        purl = "pkg:rpm/redhat/buildah@1.19.9-1.module%2Bel8.4.0%2B21078%2Ba96cfbf6?arch=src"
        cpes = {"AppStream-8.4.0.Z.TUS": cpe}
        purls = {"buildah-0:1.19.9-1.module+el8.4.0+21078+a96cfbf6.src": purl}
        result = Remediation(
            "AppStream-8.4.0.Z.TUS:container-tools:3.0:8040020240104111259:c0c392d5"
            ":buildah-0:1.19.9-1.module+el8.4.0+21078+a96cfbf6.src", cpes,
            purls)
        self.assertEqual(result.cpe, cpe)
        self.assertEqual(result.purl, purl)


if __name__ == '__main__':
    unittest.main()



"""Functions to help find the first package version for a specific release"""
import argparse
import os
import typing
import urllib.error
from datetime import datetime, timedelta
import json
import gzip
from urllib import request

import pandas as pd

DEBIAN_RELEASE_VERSIONS_URL = \
  'https://debian.pages.debian.net/distro-info-data/debian.csv'
DEBIAN_SNAPSHOT_URL = 'https://snapshot.debian.org/archive/debian/{date}/dists/'
# `.gz` format always exist for all snapshots
DEBIAN_SOURCES_URL_EXTENSION = '{version}/main/source/Sources.gz'

# List of ignored versions, mostly too early to be in snapshots
IGNORED_DEBIAN_VERSIONS = frozenset(
    ['experimental', 'buzz', 'rex', 'bo', 'hamm', 'slink', 'potato'])

# Number of days to search (day by day) if the initial date returns 404
FIRST_RELEASE_LOOKAHEAD_DAYS = 10

# First snapshot date for Debian
FIRST_SNAPSHOT_DATE = datetime(2005, 3, 12)

# Prefixes used in the Sources file
PACKAGE_KEY = 'Package: '
VERSION_KEY = 'Version: '


def get_debian_dists_url(date: datetime):
  """Create an url for snapshot.debian.org to get distribution"""
  formatted_date = convert_datetime_to_str_datetime(date)
  return DEBIAN_SNAPSHOT_URL.format(date=formatted_date)


def get_debian_sources_url(date: datetime, version: str):
  """Create an url for snapshot.debian.org"""
  formatted_date = convert_datetime_to_str_datetime(date)

  return DEBIAN_SNAPSHOT_URL.format(
      date=formatted_date) + DEBIAN_SOURCES_URL_EXTENSION.format(
          version=version)


def convert_datetime_to_str_datetime(input_datetime: datetime) -> str:
  """Convert datetime object to debian snapshot url string"""
  return input_datetime.isoformat().replace('-', '').replace(':', '') + 'Z'


def retrieve_codename_to_version() -> pd.DataFrame:
  """Returns the codename to version mapping"""
  with request.urlopen(DEBIAN_RELEASE_VERSIONS_URL) as csv:
    df = pd.read_csv(csv, dtype=str)
    # `series` appears to be `codename` but with no caps
    df['sources'] = ''
    df.dropna(subset=['release'], inplace=True)
    codename_to_version = df.set_index('series')

  return codename_to_version


def parse_created_dates_and_set_time(date: str) -> datetime:
  """Parse created date in debian version csv to datetime plus one day"""
  result = datetime.strptime(date, '%Y-%m-%d') + timedelta(days=1)
  # Set minimum date to first debian snapshot
  return max(result, FIRST_SNAPSHOT_DATE)


def load_sources(date: datetime, dist: str) -> typing.Dict[str, str]:
  """Load the sources file and store in a dictionary of {name: version}"""
  with request.urlopen(get_debian_sources_url(date, dist)) as res:
    decompressed = gzip.decompress(res.read()).decode('iso-8859-2')
    package_version_dict = {}
    current_package = None
    for line in decompressed.splitlines():
      if line.startswith(PACKAGE_KEY):
        current_package = line[len(PACKAGE_KEY):]
        continue

      if line.startswith(VERSION_KEY):
        package_version_dict[current_package] = line[len(VERSION_KEY):]
        continue

    return package_version_dict


def load_first_packages() -> pd.DataFrame:
  """Loads the dataframe containing the first version of packages per distro"""
  codename_to_version: pd.DataFrame = retrieve_codename_to_version()

  first_release_dates = zip(
      codename_to_version.index,
      codename_to_version['release'].map(parse_created_dates_and_set_time))

  first_release_dict: typing.Dict[str, datetime] = dict(
      x for x in first_release_dates if x[0] not in IGNORED_DEBIAN_VERSIONS)

  for version, date in first_release_dict.items():
    # retry for n days into the future if the first request doesn't work
    for i in range(FIRST_RELEASE_LOOKAHEAD_DAYS + 1):
      actual_date = date + timedelta(days=i)
      try:
        codename_to_version.loc[version].sources = load_sources(
            actual_date, version)
        break
      except urllib.error.HTTPError as http_error:
        if http_error.code != 404:
          raise
        # Expect 404 errors for releases before snapshot exists

      if actual_date > datetime.utcnow():
        # No need to keep trying future dates
        break

  return codename_to_version


def get_first_package_version(first_pkg_data: pd.DataFrame, package_name: str,
                              release_name: str) -> str:
  """Get first package version"""
  try:
    return first_pkg_data.loc[release_name].sources[package_name]
  except KeyError:
    # The package is not added when the image is first seen.
    # So it is safe to return 0, indicating the earliest version
    # given by the snapshot API
    return '0'


def main():
  parser = argparse.ArgumentParser(description='Detects all first versions')
  parser.add_argument(
      '-o',
      '--output-dir',
      default='first_package_output',
      help='Output folder')
  args = parser.parse_args()

  dataframe = load_first_packages()
  version_to_sources = dict(zip(dataframe['version'], dataframe['sources']))

  os.makedirs(args.output_dir, exist_ok=True)
  for version, sources in version_to_sources.items():
    versions_path = os.path.join(args.output_dir, version + '.json')
    with open(versions_path, 'w') as handle:
      handle.write(json.dumps(sources))


if __name__ == '__main__':
  main()



#!/usr/bin/env python3
""" Convert a CSAF document to OSV format
    i.e. https://access.redhat.com/security/data/csaf/v2/advisories/2024/rhsa-2024_4546.json
"""
import argparse
import json
import sys
from datetime import datetime

import requests
from jsonschema import validate
from redhat_osv.osv import DATE_FORMAT, RedHatConverter


def main():
    """
    Given a Red Hat CSAF document, covert it to OSV. Writes the OSV file to disk at 'osv' by default
    """
    parser = argparse.ArgumentParser(description='CSAF to OSV Converter')
    parser.add_argument("csaf", metavar="FILE", help='CSAF file to process')
    parser.add_argument('--output_directory', dest='out_dir', default="osv")

    args = parser.parse_args()

    with open(args.csaf, "r", encoding="utf-8") as in_f:
        csaf_data = in_f.read()

    converter = RedHatConverter()
    osv_id, osv_data = converter.convert(csaf_data,
                                         datetime.now().strftime(DATE_FORMAT))

    if not osv_data:
        sys.exit(1)

    with open(f"{args.out_dir}/{osv_id}.json", "w", encoding="utf-8") as out_f:
        out_f.write(osv_data)


if __name__ == '__main__':
    main()
